"""MA Search module."""

from threading import Lock as _Lock

from .. import util as _util
from .. import clientweb as _web
from ..namesys import Filter as _Filter
from ..namesys import SiriusPVName as _SiriusPVName
from ..magnet import util as _mutil
from .ps_search import PSSearch as _PSSearch


class MASearch:
    """MA and PM Magnet Search class."""

    _psnames_list = list()
    _manames_list = list()
    _maname_2_psnames_dict = dict()  # magnet-excitation-ps file
    _maname_2_trim_dict = dict()
    _maname_2_modeldata_dict = dict()

    _lock = _Lock()

    @staticmethod
    def get_manames(filters=None, sorting=None):
        """Return a sorted and filtered list of all magnet names.

        This list also includes pulsed magnets (PM).
        """
        MASearch._reload_maname_2_psnames_dict()
        return _Filter.process_filters(MASearch._manames_list,
                                       filters=filters,
                                       sorting=sorting)

    @staticmethod
    def get_pwrsupply_manames(filters=None, sorting=None):
        """Return a sorted and filtered list of all pwrsupply magnet names."""
        MASearch._reload_maname_2_psnames_dict()
        ps_manames_list = list(MASearch._maname_2_psnames_dict.keys())
        return _Filter.process_filters(ps_manames_list,
                                       filters=filters,
                                       sorting=sorting)

    @staticmethod
    def get_mapositions(names=None, filters=None, sorting=None):
        """Return a sorted and filtered list of all magnet positions.

        This list also includes pulsed magnets (PM).
        """
        MASearch._reload_maname_2_model_data()
        if not names:
            names = MASearch.get_manames(filters=filters, sorting=sorting)
        return [MASearch._maname_2_modeldata_dict[mag]['pos'] for mag in names]

    @staticmethod
    def get_manicknames(names=None, filters=None, sorting=None):
        """Return a list with Magnet nicknames."""
        if not names:
            names = MASearch.get_manames(filters=filters, sorting=sorting)
        nicknames = len(names)*['']
        for i, mag in enumerate(names):
            nicknames[i] = mag.sub + ('-' + mag.idx if mag.idx else '')
        return nicknames

    @staticmethod
    def conv_maname_2_trims(maname):
        """Convert maname powersupply to its trims."""
        MASearch._reload_maname_2_psnames_dict()
        return MASearch._maname_2_trim_dict.get(maname, None)

    @staticmethod
    def conv_maname_2_magfunc(maname):
        """Return a dict mapping ps to magnet functions for given magnet ps."""
        MASearch._reload_maname_2_psnames_dict()
        ps = MASearch._maname_2_psnames_dict[maname]
        ps_types = tuple(map(_PSSearch.conv_psname_2_pstype, ps))
        ma_func = tuple(map(_PSSearch.conv_pstype_2_magfunc, ps_types))

        ret = dict()
        for i, psname in enumerate(ps):
            ret[psname] = ma_func[i]

        return ret

    @staticmethod
    def conv_maname_2_psnames(maname):
        """Return list of power supplies associated with a given magnet."""
        MASearch._reload_maname_2_psnames_dict()
        return MASearch._maname_2_psnames_dict[maname]

    @staticmethod
    def conv_psmaname_2_psnames(maname):
        """Return list of power supplies associated with a given magnet ps."""
        maname = _SiriusPVName(maname)
        ma_class = _mutil.magnet_class(maname)
        if 'dipole' == ma_class:
            if 'SI' == maname.sec:
                return ['SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2']
            elif 'BO' == maname.sec:
                return ['BO-Fam:PS-B-1', 'BO-Fam:PS-B-2']
        elif 'pulsed' == ma_class:
            return [maname.replace(':PM', ':PU')]
        return [maname.replace(':MA', ':PS')]

    @staticmethod
    def conv_psname_2_psmaname(psname):
        """Return power supply maname for a given psname."""
        MASearch._reload_maname_2_psnames_dict()
        if psname not in MASearch._psnames_list or 'DCLink' in psname:
            return None
        if 'PS-B1B2' in psname:
            return 'SI-Fam:MA-B1B2'
        elif 'BO-Fam:PS-B' in psname:
            return 'BO-Fam:MA-B'
        else:
            return psname.replace('PS', 'MA').replace('PU', 'PM')

    @staticmethod
    def conv_bbbname_2_psmanames(bbbname):
        """Return psmanames for a given bbbname."""
        manames = set()
        psnames = _PSSearch.conv_bbbname_2_psnames(bbbname)
        for psname in psnames:
            maname = MASearch.conv_psname_2_psmaname(psname[0])
            if maname is None:
                continue
            manames.add(maname)
        return list(manames)

    # --- private methods ---

    @staticmethod
    def _reload_maname_2_psnames_dict():
        """Build a dict of tuples with power supplies of each magnet."""
        with MASearch._lock:
            if MASearch._maname_2_psnames_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read magnet-excitation-ps from web server!')
            text = _web.magnets_excitation_ps_read()
            data, param_dict = _util.read_text_data(text)
            maname_2_psnames_dict = dict()
            maname_2_trim_dict = dict()
            manames_list = list()
            for datum in data:
                magnet, *psnames = datum
                magnet = _SiriusPVName(magnet)
                manames_list.append(magnet)
                maname_2_psnames_dict[magnet] = tuple(psnames)
                if 'Fam' != magnet.sub:
                    famname = magnet
                    famname = famname.replace(
                        famname.sub, 'Fam').replace('MA-', 'PS-')
                    if '-Fam:PS-Q' in famname and famname in psnames:
                        psnames.remove(famname)
                        maname = famname.replace('PS-', 'MA-')
                        if maname not in maname_2_trim_dict:
                            maname_2_trim_dict[maname] = tuple(psnames)
                        else:
                            maname_2_trim_dict[maname] += tuple(psnames)
            MASearch._maname_2_psnames_dict = maname_2_psnames_dict
            MASearch._maname_2_trim_dict = maname_2_trim_dict
            MASearch._manames_list = sorted(manames_list)
            MASearch._psnames_list = _PSSearch.get_psnames()

    @staticmethod
    def _reload_maname_2_model_data():
        """Build a dictionary of model information for each magnet."""
        with MASearch._lock:
            if MASearch._maname_2_modeldata_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read magnet model data from web server!')
            text = _web.magnets_model_data()
            data, param_dict = _util.read_text_data(text)
            maname_2_modeldata_dict = dict()
            for datum in data:
                mag, *data = datum
                mag = _SiriusPVName(mag)
                maname_2_modeldata_dict[mag] = {'pos': float(data[0])}
            MASearch._maname_2_modeldata_dict = maname_2_modeldata_dict



"""."""
from .device import Device as _Device
from .device import Devices as _Devices
from .ict import ICT


class CurrInfoTranspEff(_Device):
    """."""

    class DEVICES:
        """Devices names."""

        LI = 'LI-Glob:AP-CurrInfo'
        TB = 'TB-Glob:AP-CurrInfo'
        TS = 'TS-Glob:AP-CurrInfo'
        ALL = (LI, TB, TS, )

    _properties = ('TranspEff-Mon', )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in CurrInfoTranspEff.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=CurrInfoTranspEff._properties)

    @property
    def transpeff(self):
        """."""
        return self['TranspEff-Mon']


class CurrInfoLinear(_Devices):
    """."""

    class DEVICES:
        """Devices names."""

        LI = CurrInfoTranspEff.DEVICES.LI
        TB = CurrInfoTranspEff.DEVICES.TB
        TS = CurrInfoTranspEff.DEVICES.TS
        ALL = (LI, TB, TS, )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in CurrInfoLinear.DEVICES.ALL:
            raise NotImplementedError(devname)

        if devname == CurrInfoLinear.DEVICES.LI:
            transp = CurrInfoTranspEff(CurrInfoLinear.DEVICES.LI)
            ict1 = ICT(ICT.DEVICES.LI_1)
            ict2 = ICT(ICT.DEVICES.LI_2)
        elif devname == CurrInfoLinear.DEVICES.TB:
            transp = CurrInfoTranspEff(CurrInfoLinear.DEVICES.TB)
            ict1 = ICT(ICT.DEVICES.TB_02)
            ict2 = ICT(ICT.DEVICES.TB_04)
        elif devname == CurrInfoLinear.DEVICES.TS:
            transp = CurrInfoTranspEff(CurrInfoLinear.DEVICES.TS)
            ict1 = ICT(ICT.DEVICES.TS_01)
            ict2 = ICT(ICT.DEVICES.TS_04)

        devices = (
            transp, ict1, ict2,
        )

        # call base class constructor
        super().__init__(devname, devices)

    @property
    def transpeff(self):
        """."""
        return self.devices[0].transpeff

    @property
    def charge_ict1(self):
        """."""
        return self.devices[1].charge

    @property
    def charge_ict2(self):
        """."""
        return self.devices[2].charge

    @property
    def charge_avg_ict1(self):
        """."""
        return self.devices[1].charge_avg

    @property
    def charge_avg_ict2(self):
        """."""
        return self.devices[2].charge_avg

    @property
    def charge_max_ict1(self):
        """."""
        return self.devices[1].charge_max

    @property
    def charge_max_ict2(self):
        """."""
        return self.devices[2].charge_max

    @property
    def charge_min_ict1(self):
        """."""
        return self.devices[1].charge_min

    @property
    def charge_min_ict2(self):
        """."""
        return self.devices[2].charge_min

    @property
    def charge_std_ict1(self):
        """."""
        return self.devices[1].charge_std

    @property
    def charge_std_ict2(self):
        """."""
        return self.devices[2].charge_std

    @property
    def pulse_count_ict1(self):
        """."""
        return self.devices[1].pulse_count

    @property
    def pulse_count_ict2(self):
        """."""
        return self.devices[2].pulse_count


class CurrInfoBO(_Device):
    """."""

    DEVNAME = 'BO-Glob:AP-CurrInfo'

    _properties = (
        'Charge150MeV-Mon', 'Current150MeV-Mon',
        'Charge1GeV-Mon', 'Current1GeV-Mon',
        'Charge2GeV-Mon', 'Current2GeV-Mon',
        'Charge3GeV-Mon', 'Current3GeV-Mon',
        'IntCurrent3GeV-Mon', 'RampEff-Mon',
    )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(CurrInfoBO.DEVNAME, properties=CurrInfoBO._properties)

    @property
    def charge150mev(self):
        """."""
        return self['Charge150MeV-Mon']

    @property
    def current150mev(self):
        """."""
        return self['Current150MeV-Mon']

    @property
    def charge1gev(self):
        """."""
        return self['Charge1GeV-Mon']

    @property
    def current1gev(self):
        """."""
        return self['Current1GeV-Mon']

    @property
    def charge2gev(self):
        """."""
        return self['Charge2GeV-Mon']

    @property
    def current2gev(self):
        """."""
        return self['Current2GeV-Mon']

    @property
    def charge3gev(self):
        """."""
        return self['Charge3GeV-Mon']

    @property
    def current3gev(self):
        """."""
        return self['Current3GeV-Mon']

    @property
    def intcurrent3gev(self):
        """."""
        return self['IntCurrent3GeV-Mon']

    @property
    def rampeff(self):
        """."""
        return self['RampEff-Mon']


class CurrInfoSI(_Device):
    """."""

    DEVNAME = 'SI-Glob:AP-CurrInfo'

    _properties = (
        'Charge-Mon', 'Current-Mon',
        'InjEff-Mon', 'InjCurr-Mon', 'InjCharge-Mon',
        'Lifetime-Mon', 'LifetimeBPM-Mon',
        'StoredEBeam-Mon',
        'BufferValue-Mon', 'BufferTimestamp-Mon',
    )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(CurrInfoSI.DEVNAME, properties=CurrInfoSI._properties)

    @property
    def charge(self):
        """."""
        return self['Charge-Mon']

    @property
    def current(self):
        """."""
        return self['Current-Mon']

    @property
    def injeff(self):
        """Efficiency of last injection pulse."""
        return self['InjEff-Mon']

    @property
    def injcurr(self):
        """Current of last injection pulse."""
        return self['InjCurr-Mon']

    @property
    def injcharge(self):
        """Charge of last injection pulse."""
        return self['InjCharge-Mon']

    @property
    def lifetime(self):
        """."""
        return self['Lifetime-Mon']

    @property
    def lifetimebpm(self):
        """."""
        return self['LifetimeBPM-Mon']

    @property
    def storedbeam(self):
        """."""
        return self['StoredEBeam-Mon']


class CurrInfoAS(_Devices):
    """."""

    class DEVICES:
        """Devices names."""

        LI = CurrInfoLinear.DEVICES.LI
        TB = CurrInfoLinear.DEVICES.TB
        BO = CurrInfoBO.DEVNAME
        TS = CurrInfoLinear.DEVICES.TS
        SI = CurrInfoSI.DEVNAME
        ALL = (LI, TB, BO, TS, SI, )

    def __init__(self):
        """."""
        currinfo_li = CurrInfoLinear(CurrInfoLinear.DEVICES.LI)
        currinfo_tb = CurrInfoLinear(CurrInfoLinear.DEVICES.TB)
        currinfo_bo = CurrInfoBO()
        currinfo_ts = CurrInfoLinear(CurrInfoLinear.DEVICES.TS)
        currinfo_si = CurrInfoSI()

        devices = (
            currinfo_li, currinfo_tb, currinfo_bo,
            currinfo_ts, currinfo_si
        )

        # call base class constructor
        super().__init__('', devices)

    @property
    def li(self):
        """Return LI CurrInfo device."""
        return self.devices[0]

    @property
    def tb(self):
        """Return TB CurrInfo device."""
        return self.devices[1]

    @property
    def bo(self):
        """Return BO CurrInfo device."""
        return self.devices[2]

    @property
    def ts(self):
        """Return TS CurrInfo device."""
        return self.devices[3]

    @property
    def si(self):
        """Return SI CurrInfo device."""
        return self.devices[4]



"""Define functions used to map an epics field to a BSMP entity id.

These classes implement a command interface, that is, they have
an `execute` method.
"""

import re as _re

from ..csdev import Const as _const_psdev
from ..bsmp.constants import ConstPSBSMP as _const_psbsmp


class Function:
    """Generic function class."""

    def execute(self, value):
        """Execute function."""
        raise NotImplementedError()


class BSMPFunction(Function):
    """Used PRU controller to execute a function."""

    def __init__(self, device_ids, pru_controller, func_id, setpoints=()):
        """Get controller and bsmp function id."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.func_id = func_id
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value is None:
                self.pru_controller.exec_functions(
                    self._device_ids, self.func_id)
            else:
                self.pru_controller.exec_functions(
                    self._device_ids, self.func_id, value)


class Command(BSMPFunction):
    """Execute a Cmd type PV."""

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.exec_functions(self._device_ids, self.func_id)


class BSMPFunctionNull(BSMPFunction):
    """Do nothing."""

    def __init__(self):
        """Do nothing."""
        super().__init__(None, None, None)

    def execute(self, value=None):
        """Do nothing."""


class WfmMonAcq(Function):
    """Enable/Disable Wfm-Mon curve acquisition."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Define CurveAcq."""
        self._device_ids = device_ids
        self.enable = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_ENABLE_SCOPE)
        self.disable = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_DISABLE_SCOPE)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute Command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == 1:
                self.enable.execute()
            elif value == 0:
                self.disable.execute()


class WfmSP(Function):
    """Executes a Wfm curve write command."""

    def __init__(self, device_ids, pru_controller, setpoints=()):
        """Get pru controller."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            for dev_id in self._device_ids:
                self.pru_controller.wfmref_write(dev_id, value)


class WfmUpdate(Function):
    """Wfm Update command."""

    def __init__(self, device_ids, pru_controller, setpoints=()):
        """Get controller and bsmp function id."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.wfm_update(self._device_ids, interval=0)


class WfmUpdateAutoSel(Function):
    """Enable/Disable curves acquisition."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Define CurveAcq."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute Command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == 1:
                self.pru_controller.scope_update_auto_enable()
            elif value == 0:
                self.pru_controller.scope_update_auto_disable()


class PRUCProperty(Function):
    """Executes a PRUCProperty command."""

    def __init__(self, pru_controller, property_name, setpoints=()):
        """Get pru controller."""
        self.pru_controller = pru_controller
        self.property = property_name
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            setattr(self.pru_controller, self.property, value)


class PSPwrState(Function):
    """Adapter to deal with FBP turn on/off functions."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Define function."""
        self._device_ids = device_ids
        self.turn_on = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_TURN_ON)
        self.close_loop = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CLOSE_LOOP)
        self.turn_off = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_TURN_OFF)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute Command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == 1:
                self.turn_on.execute()
                # Not necessary anymore from firmware v1.25 onwards
                # self.close_loop.execute()
            elif value == 0:
                self.turn_off.execute()


class PSPwrStateFBP_DCLink(Function):
    """Adapter to deal with FBP_DCLink turn on/off functions."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Define function."""
        self.setpoints = setpoints
        self.turn_on = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_TURN_ON)
        self.turn_off = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_TURN_OFF)
        self.open_loop = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_OPEN_LOOP)

    def execute(self, value=None):
        """Execute Command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == _const_psdev.PwrStateSel.On:
                self.turn_on.execute()
            elif value == _const_psdev.PwrStateSel.Off:
                self.turn_off.execute()


# NOTE: writers below may be implemented as a general class
# with the BMSP Function ID as an argument.


class CtrlLoop(Function):
    """Adapter to close or open control loops."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Define function."""
        self.pru_controller = pru_controller
        self.setpoints = setpoints
        self.open_loop = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_OPEN_LOOP)
        self.close_loop = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CLOSE_LOOP)

    def execute(self, value=None):
        """Execute Command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == 1:
                self.open_loop.execute(None)
            elif value == 0:
                self.close_loop.execute(None)


class PSOpMode(Function):
    """Decorate a function."""

    def __init__(self, device_ids, function, setpoints=None):
        """Command."""
        self._device_ids = device_ids
        self.function = function
        self.setpoints = setpoints

    def execute(self, value=None):
        """Parse before executing."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            if value == _const_psdev.OpMode.SlowRef:
                op_mode = _const_psdev.States.SlowRef
            elif value == _const_psdev.OpMode.SlowRefSync:
                op_mode = _const_psdev.States.SlowRefSync
            elif value == _const_psdev.OpMode.Cycle:
                op_mode = _const_psdev.States.Cycle
            elif value == _const_psdev.OpMode.RmpWfm:
                op_mode = _const_psdev.States.RmpWfm
            elif value == _const_psdev.OpMode.MigWfm:
                op_mode = _const_psdev.States.MigWfm
            else:
                op_mode = value
            # execute opmode change
            self.function.execute(op_mode)


class Current(Function):
    """Writer to set power supplies current."""
    # NOTE: This was needed when PRU was used to control timing.
    # It may be discarded...
    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Create command to set current."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.set_current = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_SET_SLOWREF)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.set_current.execute(value)


class Voltage(Function):
    """Command to set voltage in DCLink type PS."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Create command to set voltage."""
        self._device_ids = device_ids
        func_id = _const_psbsmp.F_SET_SLOWREF
        self.set_voltage = BSMPFunction(device_ids, pru_controller, func_id)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.set_voltage.execute()


class ScopeSrcAddr(Function):
    """Writer to set scope source address."""
    def __init__(self, device_ids, pru_controller, setpoints=None):
        """."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.set_address = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CFG_SOURCE_SCOPE)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.set_address.execute(value)


class ScopeFreq(Function):
    """Writer to set scpope frequency."""
    def __init__(self, device_ids, pru_controller, setpoints=None):
        """."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.set_frequency = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CFG_FREQ_SCOPE)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.set_frequency.execute(value)


class ScopeDuration(Function):
    """Writer to set scpope frequency."""
    def __init__(self, device_ids, pru_controller, setpoints=None):
        """."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.set_duration = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CFG_DURATION_SCOPE)
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.set_duration.execute(value)


class CfgSiggen(Function):
    """Command to configure siggen."""

    def __init__(self, device_ids, pru_controller, idx, setpoints=None):
        """Init."""
        self._idx = idx
        self._setpoints = setpoints
        self._cfg = BSMPFunction(
            device_ids, pru_controller, _const_psbsmp.F_CFG_SIGGEN)

    def execute(self, value=None):
        """Execute command."""
        if self._idx == 5:
            if not self._setpoints or \
                    (self._setpoints and
                     self._setpoints.apply(value[self._idx:])):
                self._cfg.execute(value)
        else:
            if not self._setpoints or \
                    (self._setpoints and
                     self._setpoints.apply(value[self._idx])):
                self._cfg.execute(value)


class SOFBCurrent(Function):
    """."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Create command to set SOFBCurrent."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.sofb_current_set(value)


class SOFBMode(Function):
    """."""

    def __init__(self, pru_controller, setpoints=None):
        """Create command to set SOFBMode."""
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.sofb_mode_set(value)


class SOFBUpdate(Function):
    """."""

    def __init__(self, pru_controller, setpoints=None):
        """Create command to set SOFBMode."""
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.sofb_update_variables_state()


class ParamUpdate(Function):
    """ParamUpdate Function."""

    def __init__(self, device_ids, pru_controller, setpoints=None):
        """Create command to update parameters."""
        self._device_ids = device_ids
        self.pru_controller = pru_controller
        self.setpoints = setpoints

    def execute(self, value=None):
        """Execute command."""
        if not self.setpoints or \
                (self.setpoints and self.setpoints.apply(value)):
            self.pru_controller.update_parameters(self._device_ids)


class Setpoint:
    """Setpoint."""

    _regexp_setpoint = _re.compile('^.*-(SP|Sel|Cmd)$')

    def __init__(self, epics_field, epics_database):
        """Init."""
        self.field = epics_field
        self.value = epics_database['value']
        self.database = epics_database
        if '-Cmd' in epics_field:
            self.is_cmd = True
        else:
            self.is_cmd = False
        self.type = epics_database['type']
        if 'count' in epics_database:
            self.count = epics_database['count']
        else:
            self.count = None
        if self.type == 'enum' and 'enums' in epics_database:
            self.enums = epics_database['enums']
        else:
            self.enums = None
        self.value = epics_database['value']
        if self.type in ('int', 'float'):
            if 'hihi' in epics_database:
                self.high = epics_database['hihi']
            else:
                self.high = None
            if 'lolo' in epics_database:
                self.low = epics_database['lolo']
            else:
                self.low = None
        else:
            self.low = None
            self.high = None

    def apply(self, value):
        """Apply setpoint value."""
        if self._check(value):
            if self.is_cmd:
                self.value += 1
            else:
                self.value = value
            return True
        return False

    def read(self):
        """Read setpoint value."""
        return self.value

    def _check(self, value):
        """Check value."""
        if self.is_cmd:
            if value > 0:
                return True
        elif self.type in ('int', 'float'):
            if self.low is None and self.high is None:
                return True
            if value is not None and \
               (value >= self.low and value <= self.high):
                return True
        elif self.type == 'enum':
            if value in tuple(range(len(self.enums))):
                return True
        return False

    @staticmethod
    def match(field):
        """Check if field is a setpoint."""
        return Setpoint._regexp_setpoint.match(field)


class Setpoints:
    """Setpoints."""

    def __init__(self, setpoints):
        """Constructor."""
        self._setpoints = setpoints

    def apply(self, value):
        """Apply setpoint."""
        for setpoint in self._setpoints:
            if not setpoint.apply(value):
                return False

        return True



"""DVF devices."""

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple
from mathphys.imgproc import Image2D_Fit as _Image2D_Fit
from mathphys.imgproc import FitGaussianScipy as _FitGaussianScipy

from .device import DeviceNC as _DeviceNC


class DVF(_DeviceNC):
    """Beam Visualization Device ("Dispositivo de Visualização de Feixe")."""

    class DEVICES:
        """Devices names."""
        CAX_DVF1 = 'CAX:A:BASLER01'
        CAX_DVF2 = 'CAX:B:BASLER01'
        ALL = (CAX_DVF1, CAX_DVF2)

    _default_timeout = 10  # [s]

    _dvfparam_fields = (
        'MAX_INTENSITY_NR_BITS',
        'ACQUISITION_TIME_MIN',  # [s]
        'ACQUISITION_TIME_DEFAULT',  # [s]
        'EXPOSURE_TIME_DEFAULT',  # [s]
        'IMAGE_SIZE_Y',  # [pixel]
        'IMAGE_SIZE_X',  # [pixel]
        'IMAGE_PIXEL_SIZE',  # [um]
        'OPTICS_MAGNIFICATION_FACTOR',  # source to image
    )

    _dev2params = {
        DEVICES.CAX_DVF1:
            _get_namedtuple(
                'DVFParameters',
                _dvfparam_fields, (16, 0.5, 0.5, 0.005, 2064, 3088, 2.4, 5.0)),
        DEVICES.CAX_DVF2:
            _get_namedtuple(
                'DVFParameters',
                _dvfparam_fields, (16, 0.5, 0.5, 0.005, 2064, 3088, 2.4, 5.0)),
        }

    _properties = (
        'cam1:MaxSizeX_RBV', 'cam1:MaxSizeY_RBV',
        'cam1:ArrayCallbacks', 'cam1:ArrayCallbacks_RBV',
        'cam1:AcquireTime', 'cam1:AcquireTime_RBV',
        'cam1:AcquirePeriod', 'cam1:AcquirePeriod_RBV',
        'cam1:Acquire', 'cam1:Acquire_RBV',
        'cam1:ImageMode', 'cam1:ImageMode_RBV',
        'cam1:Gain', 'cam1:Gain_RBV',
        'cam1:GainAuto', 'cam1:GainAuto_RBV',
        'cam1:DataType', 'cam1:DataType_RBV',
        'cam1:PixelFormat', 'cam1:PixelFormat_RBV',
        'cam1:PixelSize', 'cam1:PixelSize_RBV',
        'cam1:SizeX_RBV', 'cam1:SizeY_RBV',
        'cam1:Temperature',
        'cam1:FAILURES_RBV', 'cam1:COMPLETED_RBV',
        'ROI1:NDArrayPort', 'ROI1:NDArrayPort_RBV',
        'ROI1:EnableCallbacks', 'ROI1:EnableCallbacks_RBV',
        'ROI1:MinX', 'ROI1:MinX_RBV',
        'ROI1:MinY', 'ROI1:MinY_RBV',
        'ROI1:SizeX', 'ROI1:SizeX_RBV',
        'ROI1:SizeY', 'ROI1:SizeY_RBV',
        'ROI1:EnableX', 'ROI1:EnableX_RBV',
        'ROI1:EnableY', 'ROI1:EnableY_RBV',
        'ROI1:ArrayCallbacks', 'ROI1:ArrayCallbacks_RBV',
        'image1:NDArrayPort', 'image1:NDArrayPort_RBV',
        'image1:EnableCallbacks', 'image1:EnableCallbacks_RBV',
        'image1:ArraySize0_RBV', 'image1:ArraySize1_RBV',
        'image1:ArrayData',
        'ffmstream1:EnableCallbacks', 'ffmstream1:EnableCallbacks_RBV',
        'Trans1:EnableCallbacks', 'Trans1:EnableCallbacks_RBV',
        'HDF1:EnableCallbacks', 'HDF1:EnableCallbacks_RBV',
        )

    def __init__(self, devname, *args, **kwargs):
        """Init."""
        # check if device exists
        if devname not in DVF.DEVICES.ALL:
            raise NotImplementedError(devname)
        # call base class constructor
        super().__init__(devname, properties=self._properties, *args, **kwargs)

    @property
    def parameters(self):
        """Return DVF parameters."""
        return DVF._dev2params[self.devname]

    @property
    def intensity_saturation_value(self):
        """Image intensity saturation value."""
        pixel_format_to_num_bits = {0: 8, 1: 12}  # 0: 'Mono8, 1: 'Mono12
        data_type_to_num_bits = {0: 8, 1: 16} # 0: 'UInt8', 1: 'UInt16'
        used_bits = pixel_format_to_num_bits[self.pixel_format]
        max_bits = data_type_to_num_bits[self.data_type]
        max_intensity = (1 << used_bits) - 1
        if used_bits <= max_bits:
            # Shift intensity to align with most-significant bits
            max_intensity <<= max_bits - used_bits
        elif used_bits > max_bits:
            # Clip to maximum value allowed by the data type
            max_intensity = (1 << max_bits) - 1
        return max_intensity

    @property
    def exposure_time(self):
        """Camera exposure time [s]."""
        return self['cam1:AcquireTime']

    @exposure_time.setter
    def exposure_time(self, value):
        """Camera exposure time [s]."""
        value = max(value, 0)
        self['cam1:AcquireTime'] = value

    @property
    def acquisition_time(self):
        """Camera acquisition time [s]."""
        return self['cam1:AcquirePeriod']

    @acquisition_time.setter
    def acquisition_time(self, value):
        """Camera acquisition time [s]."""
        params = self.parameters
        value = max(value, params.ACQUISITION_TIME_MIN)
        self['cam1:AcquirePeriod'] = value

    @property
    def acquisition_status(self):
        """Return IOC acquisition status."""
        return self['cam1:Acquire']

    @property
    def cam_max_sizex(self):
        """Camera max second dimension size (pixels)."""
        return self['cam1:MaxSizeX_RBV']

    @property
    def cam_max_sizey(self):
        """Camera max first dimension size (pixels)."""
        return self['cam1:MaxSizeY_RBV']

    @property
    def cam_sizex(self):
        """Camera second dimension size (pixels)."""
        return self['cam1:SizeX_RBV']

    @property
    def cam_sizey(self):
        """Camera first dimension size (pixels)."""
        return self['cam1:SizeY_RBV']

    @property
    def roi_minx(self):
        """."""
        return self['ROI1:MinX_RBV']

    @roi_minx.setter
    def roi_minx(self, value):
        """."""
        self['ROI1:MinX'] = int(value)

    @property
    def roi_miny(self):
        """."""
        return self['ROI1:MinY_RBV']

    @roi_miny.setter
    def roi_miny(self, value):
        """."""
        self['ROI1:MinY'] = int(value)

    @property
    def image_sizex(self):
        """Image second dimension size (pixels)."""
        return self['image1:ArraySize0_RBV']

    @property
    def image_sizey(self):
        """Image first dimension size (pixels)."""
        return self['image1:ArraySize1_RBV']

    @property
    def image(self):
        """Return DVF image formatted as a (sizey, sizex) numpy matrix."""
        shape = (self.image_sizey, self.image_sizex)
        data = self['image1:ArrayData']
        image = _np.reshape(data, shape)
        return image

    @property
    def image_pixel_size(self):
        """Image pixel size [um]."""
        params = self.parameters
        return params.IMAGE_PIXEL_SIZE

    @property
    def optics_magnification_factor(self):
        """Source to image magnification factor."""
        params = self.parameters
        return params.OPTICS_MAGNIFICATION_FACTOR

    @property
    def conv_pixel_2_srcsize(self):
        """Pixel to source size convertion factor."""
        pixel_size = self.image_pixel_size
        mag_factor = self.optics_magnefication_factor
        pixel2srcsize = pixel_size / mag_factor
        return pixel2srcsize

    @property
    def gain(self):
        """Return camera gain."""
        return self['cam1:Gain_RBV']

    @gain.setter
    def gain(self, value):
        """Set camera gain."""
        self['cam1:Gain'] = value

    @property
    def gain_auto(self):
        """Return camera gain auto."""
        return self['cam1:GainAuto_RBV']

    @gain.setter
    def gain_auto(self, value):
        """Set camera gain auto."""
        self['cam1:GainAuto'] = value

    @property
    def pixel_format(self):
        """Return camera pixel format."""
        return self['cam1:PixelFormat_RBV']

    @pixel_format.setter
    def pixel_format(self, value):
        """Set camera pixel format."""
        self['cam1:PixelFormat'] = value

    @property
    def data_type(self):
        """Return camera data type."""
        return self['cam1:DataType_RBV']

    @data_type.setter
    def data_type(self, value):
        """Set camera data type."""
        self['cam1:DataType'] = value

    @property
    def pixel_size(self):
        """Return camera pixel size."""
        return self['cam1:PixelSize_RBV']

    @pixel_size.setter
    def pixel_size(self, value):
        """Set camera pixel size."""
        self['cam1:PixelSize'] = value

    @property
    def cam_temperature(self):
        """Return camera temperature"""
        return self['cam1:Temperature']

    @property
    def cam_frames_completed(self):
        """Return number of acquisition frames completed."""
        return self['cam1:COMPLETED_RBV']

    @property
    def cam_frames_failures(self):
        """Return number of acquisition frames failures."""
        return self['cam1:FAILURES_RBV']

    def cmd_reset(self, timeout=None):
        """Reset DVF to a standard configuration."""
        props_values = {
            'cam1:ArrayCallbacks': 1,  # Enable passing array
            'cam1:ImageMode': 2,  # Continuous
            'cam1:PixelFormat': 1,  # Mono12
            'cam1:DataType': 1,  # UInt16 (maybe unnecessary)
            'ROI1:NDArrayPort': 'CAMPORT',  # Take img from camport
            'ROI1:EnableCallbacks': 1,  # Enable getting from NDArrayPort
            'ROI1:MinX': 0,  # [pixel]
            'ROI1:MinY': 0,  # [pixel]
            'ROI1:SizeX': self.cam_max_sizex,  # [pixel]
            'ROI1:SizeY': self.cam_max_sizey,  # [pixel]
            'ROI1:EnableX': 1,  # Enable
            'ROI1:EnableY': 1,  # Enable
            'ROI1:ArrayCallbacks': 1,  # Enable passing array
            'image1:NDArrayPort': 'ROI1',  # image1 takes img from ROI1
            'image1:EnableCallbacks': 1,  # Enable
            'ffmstream1:EnableCallbacks': 0,  # Disable
            'HDF1:EnableCallbacks': 0,  # Disable
            'Trans1:EnableCallbacks': 0,  # Disable
        }

        # set properties
        for propty, value in props_values.items():
            self[propty] = value

        # check readback values
        for propty, value in props_values.items():
            if not self._wait(propty + '_RBV', value, timeout=timeout):
                return False

        # configure image acquisition parameters
        params = self.parameters
        self.exposure_time = params.EXPOSURE_TIME_DEFAULT
        self.acquisition_time = params.ACQUISITION_TIME_DEFAULT
        return True

    def cmd_acquire_on(self, timeout=None):
        """Tune IOC image acquisition on."""
        return self._set_and_wait('cam1:Acquire', 1, timeout=timeout)

    def cmd_acquire_off(self, timeout=None):
        """Tune IOC image acquisition off."""
        return self._set_and_wait('cam1:Acquire', 0, timeout=timeout)

    @staticmethod
    def conv_devname2parameters(devname):
        """."""
        return DVF._dev2params[devname]

    def _set_and_wait(self, propty, value, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        self[propty] = value
        return self._wait(propty + '_RBV', value, timeout=timeout)


class DVFImgProc(DVF):
    """."""

    _properties = DVF._properties + (
        'ImgIntensityMax-Mon', 'ImgIntensityMin-Mon',
        'ImgIntensitySum-Mon', 'ImgIsSaturated-Mon',
        'ImgIsWithBeam-Mon',
        'ImgIsWithBeamThreshold-SP', 'ImgIsWithBeamThreshold-RB',

        'ImgROIX-RB', 'ImgROIX-SP',
        'ImgROIXCenter-Mon', 'ImgROIXFWHM-Mon',
        'ImgROIY-RB', 'ImgROIY-SP',
        'ImgROIYCenter-Mon', 'ImgROIYFWHM-Mon',

        'ImgLog-Mon',
        'ImgROIUpdateWithFWHM-Sel', 'ImgROIUpdateWithFWHM-Sts',
        'ImgROIYUpdateWithFWHMFactor-RB', 'ImgROIYUpdateWithFWHMFactor-SP',
        'ImgROIXUpdateWithFWHMFactor-RB', 'ImgROIXUpdateWithFWHMFactor-SP',

        'ImgROIXFitMean-Mon', 'ImgROIXFitSigma-Mon',
        'ImgROIXFitAmplitude-Mon', 'ImgROIXFitError-Mon',
        'ImgROIYFitMean-Mon', 'ImgROIYFitSigma-Mon',
        'ImgROIYFitAmplitude-Mon', 'ImgROIYFitError-Mon',
        'ImgFitAngle-Mon',
        'ImgFitSigma1-Mon', 'ImgFitSigma2-Mon',
        'ImgFitProcTime-Mon',
        'ImgFitAngleUseCMomSVD-Sel', 'ImgFitAngleUseCMomSVD-Sts',
        'ImgDVFStatus-Mon', 'ImgDVFStatusLabels-Cte',
        )

    def __init__(self, devname, *args, **kwargs):
        """."""
        super().__init__(devname=devname, *args, **kwargs)
        self._fitgaussian = _FitGaussianScipy()

    @property
    def intensity_min(self):
        """Image min intensity."""
        return self['ImgIntensityMin-Mon']

    @property
    def intensity_max(self):
        """Image max intensity."""
        return self['ImgIntensityMax-Mon']

    @property
    def intensity_sum(self):
        """Image sum intensity."""
        return self['ImgIntensitySum-Mon']

    @property
    def is_saturated(self):
        """Whether image is saturated."""
        return self['ImgIsSaturated-Mon']

    @property
    def is_with_beam(self):
        """Whether image is with beam."""
        return self['ImgIsWithBeam-Mon']

    @property
    def is_with_beam_threashold(self):
        """Get image is with beam threashold."""
        return self['ImgIsWithBeamThreshold-RB']

    @is_with_beam_threashold.setter
    def is_with_beam_threashold(self, value):
        """Set image is with beam threashold."""
        self['ImgIsWithBeamThreshold-SP'] = value

    @property
    def roiy(self):
        """."""
        return self['ImgROIY-RB']

    @roiy.setter
    def roiy(self, value):
        """."""
        self['ImgROIY-SP'] = value

    @property
    def roix(self):
        """."""
        return self['ImgROIX-RB']

    @roix.setter
    def roix(self, value):
        """."""
        self['ImgROIX-SP'] = value

    @property
    def roiy_center(self):
        """."""
        return self['ImgROIYCenter-Mon']

    @property
    def roix_center(self):
        """."""
        return self['ImgROIXCenter-Mon']

    @property
    def roiy_fwhm(self):
        """."""
        return self['ImgROIYFWHM-Mon']

    @property
    def roix_fwhm(self):
        """."""
        return self['ImgROIXFWHM-Mon']

    @property
    def roiy_fit_amplitude(self):
        """."""
        return self['ImgROIYFitAmplitude-Mon']

    @property
    def roix_fit_amplitude(self):
        """."""
        return self['ImgROIXFitAmplitude-Mon']

    @property
    def roiy_fit_mean(self):
        """."""
        return self['ImgROIYFitMean-Mon']

    @property
    def roix_fit_mean(self):
        """."""
        return self['ImgROIXFitMean-Mon']

    @property
    def roiy_fit_sigma(self):
        """."""
        return self['ImgROIYFitSigma-Mon']

    @property
    def roix_fit_sigma(self):
        """."""
        return self['ImgROIXFitSigma-Mon']

    @property
    def roiy_fit_error(self):
        """."""
        return self['ImgROIYFitError-Mon']

    @property
    def roix_fit_error(self):
        """."""
        return self['ImgROIXFitError-Mon']

    @property
    def roiy_fwhm_factor(self):
        """."""
        return self['ImgROIYUpdateWithFWHMFactor-RB']

    @roiy_fwhm_factor.setter
    def roiy_fwhm_factor(self, value):
        """."""
        self['ImgROIYUpdateWithFWHMFactor-SP'] = value

    @property
    def roix_fwhm_factor(self):
        """."""
        return self['ImgROIXUpdateWithFWHMFactor-RB']

    @roix_fwhm_factor.setter
    def roix_fwhm_factor(self, value):
        """."""
        self['ImgROIXUpdateWithFWHMFactor-SP'] = value

    @property
    def roi_update_with_fwhm(self):
        """."""
        return bool(self['ImgROIUpdateWithFWHM-Sts'])

    @roi_update_with_fwhm.setter
    def roi_update_with_fwhm(self, value):
        """."""
        self['ImgROIUpdateWithFWHM-Sel'] = bool(value)

    @property
    def fit_angle(self):
        """."""
        return self['ImgFitAngle-Mon']

    @property
    def fit_sigma1(self):
        """."""
        return self['ImgFitSigma1-Mon']

    @property
    def fit_sigma2(self):
        """."""
        return self['ImgFitSigma2-Mon']

    @property
    def fit_proctime(self):
        """Return image processing time [ms]."""
        return self['ImgFitProcTime-Mon']

    @property
    def fit_angle_use_cmom_svd(self):
        """."""
        return self['ImgFitAngleUseCMomSVD-Sts']

    @fit_angle_use_cmom_svd.setter
    def fit_angle_use_cmom_svd(self, value):
        """."""
        self['ImgFitAngleUseCMomSVD-Sel'] = bool(value)

    @property
    def log(self):
        """."""
        return self['ImgLog-Mon']

    def create_image2dfit(self):
        """Return a Image2DFit object with current image as data."""
        imgfit2d = _Image2D_Fit(
            data=self.image, fitgauss=self._fitgaussian,
            roix=self.roix, roiy=self.roiy)
        return imgfit2d



"""Define PVs, constants and properties of CurrInfo SoftIOCs."""
from copy import deepcopy as _dcopy

import numpy as _np

from .. import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    DCCTSELECTIONTYP = ('DCCT13C4', 'DCCT14C4')
    BUFFAUTORSTTYP = ('Off', 'DCurrCheck')
    FITTYP = ('Exponential', 'Linear')


_et = ETypes  # syntactic sugar


# --- Const class ---

class Const(_csdev.Const):
    """Const class defining CurrInfo constants and Enum types."""

    DCCT = _csdev.Const.register('DCCT', _et.DCCTSELECTIONTYP)
    DCCTFltCheck = _csdev.Const.register('DCCTFltCheck', _et.OFF_ON)
    BuffAutoRst = _csdev.Const.register('BuffAutoRst', _et.BUFFAUTORSTTYP)
    Fit = _csdev.Const.register('Fit', _et.FITTYP)


_c = Const  # syntactic sugar


# --- Databases ---

def get_currinfo_database(acc):
    """Return CurrentInfo Soft IOC database."""
    acc = acc.upper()
    if acc == 'TS':
        pvs_db = get_litbts_currinfo_database(acc)
    elif acc == 'LI':
        pvs_db = get_litbts_currinfo_database('LI')
        pvs_db.update(get_litbts_currinfo_database('TB'))
    elif acc == 'SI':
        pvs_db = get_si_currinfo_database()
    elif acc == 'BO':
        pvs_db = get_bo_currinfo_database()
    return pvs_db


def get_litbts_currinfo_database(acc):
    """Return LI, TB, TS, CurrentInfo Soft IOC database."""
    pref = acc + '-Glob:AP-CurrInfo:'
    pvs_db = {pref + 'Version-Cte': {'type': 'string', 'value': 'UNDEF'}}

    devices = {
        'LI': ('LI-01:DI-ICT-1:', 'LI-01:DI-ICT-2:'),
        'TB': ('TB-02:DI-ICT:', 'TB-04:DI-ICT:'),
        'TS': ('TS-01:DI-ICT:', 'TS-04:DI-ICT:')}
    def_db = {'type': 'float', 'value': 0.0, 'unit': 'nC', 'prec': 4}
    pvs = [
        'Charge-Mon', 'ChargeAvg-Mon', 'ChargeMin-Mon', 'ChargeMax-Mon',
        'ChargeStd-Mon']

    for device in devices[acc]:
        pvs_db.update({device+pv: _dcopy(def_db) for pv in pvs})
        pvs_db[device+'PulseCount-Mon'] = {'type': 'int', 'value': 0}

    def_db = {'type': 'float', 'value': 0.0, 'unit': '%', 'prec': 2}
    pvs_db[pref + 'TranspEff-Mon'] = _dcopy(def_db)
    pvs_db[pref + 'TranspEffAvg-Mon'] = _dcopy(def_db)
    pvs_db = _csdev.add_pvslist_cte(pvs_db, prefix=pref)
    return pvs_db


def get_bo_currinfo_database():
    """Return BO CurrentInfo Soft IOC database."""
    pvs_db = {'Version-Cte': {'type': 'string', 'value': 'UNDEF'}}

    pvs_db['RawReadings-Mon'] = {
        'type': 'float', 'count': 100000, 'value': _np.array(100000*[0.0]),
        'unit': 'mA'}
    pvs_db['Current150MeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA'}
    pvs_db['Current1GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA'}
    pvs_db['Current2GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA'}
    pvs_db['Current3GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA'}

    pvs_db['Charge150MeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 4, 'unit': 'nC'}
    pvs_db['Charge1GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 4, 'unit': 'nC'}
    pvs_db['Charge2GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 4, 'unit': 'nC'}
    pvs_db['Charge3GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 4, 'unit': 'nC'}

    pvs_db['IntCurrent3GeV-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 9, 'unit': 'mA.h'}

    pvs_db['CurrThold-SP'] = {
        'type': 'float', 'value': 0.004, 'prec': 4, 'unit': 'mA',
        'lolo': 0.0, 'low': 0.0, 'lolim': 0.0,
        'hilim': 0.010, 'high': 0.010, 'hihi': 0.010}
    pvs_db['CurrThold-RB'] = {
        'type': 'float', 'value': 0.004, 'prec': 4, 'unit': 'mA',
        'lolo': 0.0, 'low': 0.0, 'lolim': 0.0,
        'hilim': 0.010, 'high': 0.010, 'hihi': 0.010}

    pvs_db['RampEff-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 2, 'unit': '%'}
    pvs_db = _csdev.add_pvslist_cte(pvs_db)
    return pvs_db


def get_si_currinfo_database():
    """Return SI CurrentInfo Soft IOC database."""
    dev = 'SI-Glob:AP-CurrInfo:'
    pvs_db = {dev+'Version-Cte': {'type': 'string', 'value': 'UNDEF'}}

    pvs_db[dev+'Current-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA'}
    pvs_db[dev+'StoredEBeam-Mon'] = {'type': 'int', 'value': 0}

    pvs_db[dev+'DCCT-Sel'] = {
        'type': 'enum', 'value': _c.DCCT.DCCT13C4,
        'enums': _et.DCCTSELECTIONTYP}
    pvs_db[dev+'DCCT-Sts'] = {
        'type': 'enum', 'value': _c.DCCT.DCCT13C4,
        'enums': _et.DCCTSELECTIONTYP}

    pvs_db[dev+'DCCTFltCheck-Sel'] = {
        'type': 'enum', 'enums': _et.OFF_ON, 'value': _c.DCCTFltCheck.Off}
    pvs_db[dev+'DCCTFltCheck-Sts'] = {
        'type': 'enum', 'enums': _et.OFF_ON, 'value': _c.DCCTFltCheck.Off}

    pvs_db[dev+'Charge-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 12, 'unit': 'A.h', 'scan': 60}

    pvs_db[dev+'InjEff-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 2, 'unit': '%'}
    pvs_db[dev+'InjCurr-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 'mA'}
    pvs_db[dev+'InjCharge-Mon'] = {
        'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 'nC'}

    dev = 'AS-Glob:AP-CurrInfo:'
    pvs_db[dev+'InjCount-Mon'] = {'type': 'int', 'value': 0}

    pvs_db = _csdev.add_pvslist_cte(pvs_db, prefix=dev)
    return pvs_db


def get_lifetime_database():
    """Return CurrentInfo-Lifetime Soft IOC database."""
    pvs_db = {
        'VersionLifetime-Cte': {'type': 'string', 'value': 'UNDEF'},

        'LtFitMode-Sel': {
            'type': 'enum', 'enums': _et.FITTYP, 'value': _c.Fit.Linear},
        'LtFitMode-Sts': {
            'type': 'enum', 'enums': _et.FITTYP, 'value': _c.Fit.Linear},
        'MaxSplIntvl-SP': {
            'type': 'float', 'unit': 's', 'prec': 2, 'value': 500,
            'lolim': -1, 'hilim': 360000},
        'MaxSplIntvl-RB': {
            'type': 'float', 'unit': 's', 'prec': 2, 'value': 500,
            'lolim': -1, 'hilim': 360000},
        'SplIntvl-Mon': {
            'type': 'float', 'unit': 's', 'prec': 2, 'value': 0.0},
        'SplIntvlBPM-Mon': {
            'type': 'float', 'unit': 's', 'prec': 2, 'value': 0.0},
        'MinIntvlBtwSpl-SP': {
            'type': 'float', 'unit': 's', 'prec': 2, 'value': 0,
            'lolim': 0, 'low': 0, 'lolo': 0,
            'hilim': 1200, 'high': 1200, 'hihi': 1200},
        'MinIntvlBtwSpl-RB': {
            'type': 'float', 'value': 0, 'prec': 2, 'unit': 's'},
        'CurrOffset-SP': {
            'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA',
            'lolim': -10.0, 'low': -10.0, 'lolo': -10.0,
            'hilim': 10.0, 'high': 10.0, 'hihi': 10.0},
        'CurrOffset-RB': {
            'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'mA',
            'lolim': -10.0, 'low': -10.0, 'lolo': -10.0,
            'hilim': 10.0, 'high': 10.0, 'hihi': 10.0},

        'BuffRst-Cmd': {'type': 'int', 'value': 0},
        'BuffAutoRst-Sel': {
            'type': 'enum', 'enums': _et.BUFFAUTORSTTYP,
            'value': _c.BuffAutoRst.DCurrCheck},
        'BuffAutoRst-Sts': {
            'type': 'enum', 'enums': _et.BUFFAUTORSTTYP,
            'value': _c.BuffAutoRst.DCurrCheck},
        'BuffAutoRstDCurr-SP': {
            'type': 'float', 'value': 0.01, 'prec': 2, 'unit': 'mA',
            'lolim': -300.0, 'low': -300.0, 'lolo': -300.0,
            'hilim': 300.0, 'high': 300.0, 'hihi': 300.0},
        'BuffAutoRstDCurr-RB': {
            'type': 'float', 'value': 0.01, 'prec': 2, 'unit': 'mA',
            'lolim': -300.0, 'low': -300.0, 'lolo': -300.0,
            'hilim': 300.0, 'high': 300.0, 'hihi': 300.0},

        'FrstSplTime-SP': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},
        'FrstSplTime-RB': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},
        'LastSplTime-SP': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},
        'LastSplTime-RB': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},
        'FrstSplTimeBPM-RB': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},
        'LastSplTimeBPM-RB': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'lolim': -1.0, 'hilim': 2e10},

        'Lifetime-Mon': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'scan': 0.5},
        'BuffSize-Mon': {'type': 'int', 'value': 0},
        'BuffSizeTot-Mon': {'type': 'int', 'value': 0},
        'BufferValue-Mon': {
            'type': 'float', 'prec': 3, 'count': 100000,
            'value': [0.0, ] * 100000},
        'BufferTimestamp-Mon': {
            'type': 'float', 'prec': 3, 'count': 100000,
            'value': [0.0, ] * 100000},

        'LifetimeBPM-Mon': {
            'type': 'float', 'value': 0.0, 'prec': 2, 'unit': 's',
            'scan': 0.5},
        'BuffSizeBPM-Mon': {'type': 'int', 'value': 0},
        'BuffSizeTotBPM-Mon': {'type': 'int', 'value': 0},
        'BufferValueBPM-Mon': {
            'type': 'float', 'prec': 3, 'count': 100000,
            'value': [0.0, ] * 100000},
        'BufferTimestampBPM-Mon': {
            'type': 'float', 'prec': 3, 'count': 100000,
            'value': [0.0, ] * 100000},
        }
    pvs_db = _csdev.add_pvslist_cte(pvs_db)
    return pvs_db



"""Orbit Interlock devices."""

import numpy as _np

from ..search import BPMSearch as _BPMSearch
from ..util import ClassProperty as _classproperty
from .device import Device as _Device, Devices as _Devices


class BaseOrbitIntlk:
    """Base BPM Orbit Interlock Info."""

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    UP_2_DOWN = {
        'M1': 'M2',
        'C1-1': 'C1-2',
        'C2': 'C3-1',
        'C3-2': 'C4',
    }
    DOWN_2_UP = {val: key for key, val in UP_2_DOWN.items()}

    __BPM_NAMES = None
    __BPM_NICKNAMES = None
    __BPM_POS = None
    __BPM_DOWNSTREAM = None
    __BPM_UPSTREAM = None

    @_classproperty
    def BPM_NAMES(cls):
        """BPM names."""
        if cls.__BPM_NAMES is not None:
            return cls.__BPM_NAMES

        cls.__BPM_NAMES = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        cls.__BPM_NICKNAMES = _BPMSearch.get_nicknames(cls.__BPM_NAMES)
        cls.__BPM_POS = _BPMSearch.get_positions(cls.__BPM_NAMES)

        def _parse_nick_down_up(name, nick):
            sub, idx, *_ = nick.split('-') + ['']
            return name.substitute(sub=name.sub[:2]+sub, idx=idx)

        downnames, upnames = [], []
        for bpm in cls.__BPM_NAMES:
            nick = bpm.sub[2:]+('-' + bpm.idx if bpm.idx else '')
            if nick in BaseOrbitIntlk.DOWN_2_UP:
                down = bpm
                upnick = BaseOrbitIntlk.DOWN_2_UP[nick]
                upn = _parse_nick_down_up(bpm, upnick)
            elif nick in BaseOrbitIntlk.UP_2_DOWN:
                upn = bpm
                downnick = BaseOrbitIntlk.UP_2_DOWN[nick]
                down = _parse_nick_down_up(bpm, downnick)
            else:
                down, upn = '', ''
            downnames.append(down)
            upnames.append(upn)
        cls.__BPM_DOWNSTREAM = downnames
        cls.__BPM_UPSTREAM = upnames
        return cls.__BPM_NAMES

    @_classproperty
    def BPM_NICKNAMES(cls):
        """BPM nicknames."""
        cls.BPM_NAMES
        return cls.__BPM_NICKNAMES

    @_classproperty
    def BPM_POS(cls):
        """BPM positions."""
        cls.BPM_NAMES
        return cls.__BPM_POS

    @_classproperty
    def BPM_DOWNSTREAM(cls):
        """BPM downstream."""
        cls.BPM_NAMES
        return cls.__BPM_DOWNSTREAM

    @_classproperty
    def BPM_UPSTREAM(cls):
        """BPM upstream."""
        cls.BPM_NAMES
        return cls.__BPM_UPSTREAM

    def __init__(self):
        """Init."""
        self._oper = {
            'mean': self._mean,
            'diff': self._diff,
        }

    @staticmethod
    def get_down_up_bpms(bpmname):
        """Return down and up BPM names for bpmname."""
        index = BaseOrbitIntlk.BPM_NAMES.index(bpmname)
        down = BaseOrbitIntlk.BPM_DOWNSTREAM[index]
        upn = BaseOrbitIntlk.BPM_UPSTREAM[index]
        return down, upn

    def calc_intlk_metric(self, posarray, operation='', metric=''):
        """Return interlock metric, position or angulation."""
        if not operation:
            if not metric:
                raise ValueError(
                    'either the operation or the metric is required')
            operation = 'mean' if 'pos' in metric.lower() else 'diff'

        data_values = list()
        for bpm in BaseOrbitIntlk.BPM_NAMES:
            down, upn = BaseOrbitIntlk.get_down_up_bpms(bpm)
            dval = posarray[BaseOrbitIntlk.BPM_NAMES.index(down)]
            uval = posarray[BaseOrbitIntlk.BPM_NAMES.index(upn)]
            func = self._oper[operation]
            val = func(dval, uval)
            data_values.append(val)
        return data_values

    @staticmethod
    def _mean(var1, var2):
        return (var1 + var2)/2

    @staticmethod
    def _diff(var1, var2):
        return var1 - var2


class BPMOrbitIntlk(BaseOrbitIntlk, _Device):
    """This device group the orbit interlock  PVs from one BPM."""

    _properties = (
        # ==============================================================
        # Basic properties
        'PosX-Mon', 'PosY-Mon', 'Sum-Mon',
        # ==============================================================
        # General
        # +++++++
        # General interlock enable:
        'IntlkEn-Sel', 'IntlkEn-Sts',
        # General interlock clear:
        'IntlkClr-Cmd',  # maybe -Cmd?
        # Minimum sum threshold enable:
        # Habilita interlock de órbita apenas quando threshold da soma
        # ultrapassar o valor em "IntlkLmtMinSum-SP"
        'IntlkMinSumEn-Sel', 'IntlkMinSumEn-Sts',
        # Minimum sum threshold (em contagens da Soma da taxa FAcq):
        'IntlkLmtMinSum-SP', 'IntlkLmtMinSum-RB',
        # Status Instantâneo:
        # Interlock instântaneo, dificilmente será detectado com
        # a implementação atual do gateware
        'Intlk-Mon',
        # Latch do interlock, limpo apenas acionando-se a PV "Clr"
        # correspondente
        'IntlkLtc-Mon',
        # ===============================================================
        # Position (interlock de posição)
        # +++++++++++++++++++++++++++++++++++++
        # ***************************************************************
        # Condição para interlock de posição:
        #   thres_min > (pos BPM downstream + pos BPM upstream)/2 or
        #   thres_max < (pos BPM downstream + pos BPM upstream)/2
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ***************************************************************
        # Position interlock enable:
        'IntlkPosEn-Sel', 'IntlkPosEn-Sts',
        # Position interlock clear:
        'IntlkPosClr-Cmd',
        # Thresholds (em nm da taxa Monit1):
        'IntlkLmtPosMaxX-SP', 'IntlkLmtPosMaxX-RB',
        'IntlkLmtPosMinX-SP', 'IntlkLmtPosMinX-RB',
        'IntlkLmtPosMaxY-SP', 'IntlkLmtPosMaxY-RB',
        'IntlkLmtPosMinY-SP', 'IntlkLmtPosMinY-RB',
        # Todos os interlocks são mascarados pelo "Enable"
        # Status Instantâneo:
        'IntlkPosLower-Mon', 'IntlkPosUpper-Mon',  # X ou Y
        'IntlkPosLowerX-Mon', 'IntlkPosUpperX-Mon',  # X
        'IntlkPosLowerY-Mon', 'IntlkPosUpperY-Mon',  # Y
        # Status Latch, limpo apenas acionando-se a PV "Clr" correspondente:
        'IntlkPosLowerLtc-Mon', 'IntlkPosUpperLtc-Mon',
        'IntlkPosLowerLtcX-Mon', 'IntlkPosUpperLtcX-Mon',
        'IntlkPosLowerLtcY-Mon', 'IntlkPosUpperLtcY-Mon',
        # =============================================================
        # Angular (interlock de ângulo)
        # +++++++++++++++++++++++++++++
        # *************************************************************
        # Condição para interlock de ângulo:
        #   thres_min > (posição BPM downstream - posição BPM upstream) or
        #   thres_max < (posição BPM downstream - posição BPM upstream)
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ************************************************************
        # Angulation interlock enable:
        'IntlkAngEn-Sel', 'IntlkAngEn-Sts',
        # Angulation interlock clear:
        'IntlkAngClr-Cmd',
        # Thresholds (em rad.nm da taxa FAcq).
        # Thresholds devem ser calculados como ângulo (em rad)
        #  entre os 2 BPMs adjacentes * distância (em nm) entre eles):
        'IntlkLmtAngMaxX-SP', 'IntlkLmtAngMaxX-RB',
        'IntlkLmtAngMinX-SP', 'IntlkLmtAngMinX-RB',
        'IntlkLmtAngMaxY-SP', 'IntlkLmtAngMaxY-RB',
        'IntlkLmtAngMinY-SP', 'IntlkLmtAngMinY-RB',
        # Todos os interlocks são mascarados pelo "Enable"
        # Status Instantâneo:
        'IntlkAngLower-Mon', 'IntlkAngUpper-Mon',
        'IntlkAngLowerX-Mon', 'IntlkAngUpperX-Mon',  # X
        'IntlkAngLowerY-Mon', 'IntlkAngUpperY-Mon',  # Y
        # Status Latch, limpo apenas acionando-se a PV "Clr" correspondente:
        'IntlkAngLowerLtc-Mon', 'IntlkAngUpperLtc-Mon',
        'IntlkAngLowerLtcX-Mon', 'IntlkAngUpperLtcX-Mon',
        'IntlkAngLowerLtcY-Mon', 'IntlkAngUpperLtcY-Mon',
        # ============================================================
        )

    def __init__(self, devname):
        """Init."""
        # call base class constructor
        BaseOrbitIntlk.__init__(self)
        if devname not in self.BPM_NAMES:
            raise ValueError(devname + ' is no a valid BPM name.')
        _Device.__init__(self, devname, properties=BPMOrbitIntlk._properties)

    @property
    def posx(self):
        """Position X, Monit rate."""
        return self['PosX-Mon'] * self.CONV_NM2UM

    @property
    def posy(self):
        """Position Y, Monit rate."""
        return self['PosY-Mon'] * self.CONV_NM2UM

    @property
    def possum(self):
        """Sum, Monit rate."""
        return self['Sum-Mon']

    @property
    def pair_down_up_bpms(self):
        """Return pair of downstream and upstream BPM names."""
        return self.get_down_up_bpms(self.devname)

    # --- general interlock ---

    @property
    def gen_enable(self):
        """General interlock enable."""
        return self['IntlkEn-Sts']

    @gen_enable.setter
    def gen_enable(self, value):
        self['IntlkEn-Sel'] = int(value)

    def cmd_reset_gen(self):
        """General interlock clear."""
        self['IntlkClr-Cmd'] = 1
        return True

    @property
    def gen_inst(self):
        """Instantaneous general interlock."""
        return self['Intlk-Mon']

    @property
    def gen_latch(self):
        """Latch general interlock."""
        return self['IntlkLtc-Mon']

    # --- minimum sum threshold ---

    @property
    def minsumthres_enable(self):
        """
        Minimum sum threshold enable.

        If enabled, generate orbit interlock only when sum threshold exceeds
        value in 'minsumthres' property.
        """
        return self['IntlkMinSumEn-Sts']

    @minsumthres_enable.setter
    def minsumthres_enable(self, value):
        self['IntlkMinSumEn-Sel'] = int(value)

    @property
    def minsumthres(self):
        """Minimum sum threshold [sum count, FAcq rate]."""
        return self['IntlkLmtMinSum-RB']

    @minsumthres.setter
    def minsumthres(self, value):
        self['IntlkLmtMinSum-SP'] = int(value)

    # --- position interlock ---

    @property
    def pos_enable(self):
        """Position interlock enable."""
        return self['IntlkPosEn-Sts']

    @pos_enable.setter
    def pos_enable(self, value):
        self['IntlkPosEn-Sel'] = int(value)

    def cmd_reset_pos(self):
        """Position interlock clear."""
        self['IntlkPosClr-Cmd'] = 1
        return True

    @property
    def pos_thresminx(self):
        """Minimum X position threshold."""
        return self['IntlkLmtPosMinX-RB']

    @pos_thresminx.setter
    def pos_thresminx(self, value):
        self['IntlkLmtPosMinX-SP'] = value

    @property
    def pos_thresmaxx(self):
        """Maximum X position threshold."""
        return self['IntlkLmtPosMaxX-RB']

    @pos_thresmaxx.setter
    def pos_thresmaxx(self, value):
        self['IntlkLmtPosMaxX-SP'] = value

    @property
    def pos_thresminy(self):
        """Minimum Y position threshold."""
        return self['IntlkLmtPosMinY-RB']

    @pos_thresminy.setter
    def pos_thresminy(self, value):
        self['IntlkLmtPosMinY-SP'] = value

    @property
    def pos_thresmaxy(self):
        """Maximum Y position threshold."""
        return self['IntlkLmtPosMaxY-RB']

    @pos_thresmaxy.setter
    def pos_thresmaxy(self, value):
        self['IntlkLmtPosMaxY-SP'] = value

    @property
    def pos_inst_lower(self):
        """
        Instantaneous position interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkPosLower-Mon']

    @property
    def pos_inst_upper(self):
        """
        Instantaneous position interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkPosUpper-Mon']

    @property
    def pos_inst_lower_x(self):
        """
        Instantaneous position interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerX-Mon']

    @property
    def pos_inst_upper_x(self):
        """
        Instantaneous position interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperX-Mon']

    @property
    def pos_inst_lower_y(self):
        """
        Instantaneous position interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerY-Mon']

    @property
    def pos_inst_upper_y(self):
        """
        Instantaneous position interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperY-Mon']

    @property
    def pos_latch_lower(self):
        """
        Latch position interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkPosLowerLtc-Mon']

    @property
    def pos_latch_upper(self):
        """
        Latch position interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkPosUpperLtc-Mon']

    @property
    def pos_latch_lower_x(self):
        """
        Latch position interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerLtcX-Mon']

    @property
    def pos_latch_upper_x(self):
        """
        Latch position interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperLtcX-Mon']

    @property
    def pos_latch_lower_y(self):
        """
        Latch position interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerLtcY-Mon']

    @property
    def pos_latch_upper_y(self):
        """
        Latch position interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperLtcY-Mon']

    #  --- angulation interlock ---

    @property
    def ang_enable(self):
        """Angulation interlock enable."""
        return self['IntlkAngEn-Sts']

    @ang_enable.setter
    def ang_enable(self, value):
        self['IntlkAngEn-Sel'] = int(value)

    def cmd_reset_ang(self):
        """Angulation interlock clear."""
        self['IntlkAngClr-Cmd'] = 1
        return True

    @property
    def ang_thresminx(self):
        """Minimum X angulation threshold."""
        return self['IntlkLmtAngMinX-RB']

    @ang_thresminx.setter
    def ang_thresminx(self, value):
        self['IntlkLmtAngMinX-SP'] = value

    @property
    def ang_thresmaxx(self):
        """Maximum X angulation threshold."""
        return self['IntlkLmtAngMaxX-RB']

    @ang_thresmaxx.setter
    def ang_thresmaxx(self, value):
        self['IntlkLmtAngMaxX-SP'] = value

    @property
    def ang_thresminy(self):
        """Minimum Y angulation threshold."""
        return self['IntlkLmtAngMinY-RB']

    @ang_thresminy.setter
    def ang_thresminy(self, value):
        self['IntlkLmtAngMinY-SP'] = value

    @property
    def ang_thresmaxy(self):
        """Maximum Y angulation threshold."""
        return self['IntlkLmtAngMaxY-RB']

    @ang_thresmaxy.setter
    def ang_thresmaxy(self, value):
        self['IntlkLmtAngMaxY-SP'] = value

    @property
    def ang_inst_lower(self):
        """
        Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkAngLower-Mon']

    @property
    def ang_inst_upper(self):
        """
        Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkAngUpper-Mon']

    @property
    def ang_inst_lower_x(self):
        """
        Instantaneous angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerX-Mon']

    @property
    def ang_inst_upper_x(self):
        """
        Instantaneous angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperX-Mon']

    @property
    def ang_inst_lower_y(self):
        """
        Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerY-Mon']

    @property
    def ang_inst_upper_y(self):
        """
        Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperY-Mon']

    @property
    def ang_latch_lower(self):
        """
        Latch angulation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkAngLowerLtc-Mon']

    @property
    def ang_latch_upper(self):
        """
        Latch angulation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkAngUpperLtc-Mon']

    @property
    def ang_latch_lower_x(self):
        """
        Latch angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerLtcX-Mon']

    @property
    def ang_latch_upper_x(self):
        """
        Latch angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperLtcX-Mon']

    @property
    def ang_latch_lower_y(self):
        """
        Latch angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerLtcY-Mon']

    @property
    def ang_latch_upper_y(self):
        """
        Latch angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperLtcY-Mon']


class OrbitInterlock(BaseOrbitIntlk, _Devices):
    """Orbit Interlock device."""

    TIMEOUT = 10

    class DEVICES:
        """."""
        SI = 'SI-Fam:DI-BPM'
        ALL = (SI, )

    def __init__(self, devname=None):
        """Init."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')
        BaseOrbitIntlk.__init__(self)
        devs = [BPMOrbitIntlk(dev) for dev in self.BPM_NAMES]
        _Devices.__init__(self, devname, devs)

    # --- general interlock ---

    def cmd_gen_enable(self, timeout=TIMEOUT):
        """Enable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkEn-Sts', 1, timeout=timeout)

    def cmd_gen_disable(self, timeout=TIMEOUT):
        """Disable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkEn-Sts', 0, timeout=timeout)

    def cmd_reset_gen(self):
        """Reset all BPM general interlock."""
        for dev in self.devices:
            dev.cmd_reset_gen()
        return True

    @property
    def gen_inst(self):
        """General instantaneous interlocks.

        Returns:
            gen_inst (numpy.ndarray, 160):
                general instantaneous interlock for each BPM.
        """
        return _np.array([b.gen_inst for b in self._devices])

    @property
    def gen_latch(self):
        """General latch interlocks.

        Returns:
            gen_latch (numpy.ndarray, 160):
                general latch interlock for each BPM.
        """
        return _np.array([b.gen_latch for b in self._devices])

    # --- minimum sum threshold ---

    def cmd_minsumthres_enable(self, timeout=TIMEOUT):
        """Enable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsumthres_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkMinSumEn-Sts', 1, timeout=timeout)

    def cmd_minsumthres_disable(self, timeout=TIMEOUT):
        """Disable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsumthres_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkMinSumEn-Sts', 0, timeout=timeout)

    @property
    def minsumthres(self):
        """Minimum sum thresholds.

        Returns:
            thres (numpy.ndarray, 160): min.sum threshold for each BPM.
        """
        return _np.array([b.minsumthres for b in self._devices])

    @minsumthres.setter
    def minsumthres(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.minsumthres = value[idx]

    # --- position interlock ---

    def cmd_pos_enable(self, timeout=TIMEOUT):
        """Enable all BPM position interlock."""
        for dev in self.devices:
            dev.pos_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkPosEn-Sts', 1, timeout=timeout)

    def cmd_pos_disable(self, timeout=TIMEOUT):
        """Disable all BPM position interlock."""
        for dev in self.devices:
            dev.pos_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkPosEn-Sts', 0, timeout=timeout)

    def cmd_reset_pos(self):
        """Reset all BPM position interlock."""
        for dev in self.devices:
            dev.cmd_reset_pos()
        return True

    @property
    def pos_thresminx(self):
        """Minimum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x position threshold for each BPM.
        """
        return _np.array([b.pos_thresminx for b in self._devices])

    @pos_thresminx.setter
    def pos_thresminx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.pos_thresminx = value[idx]

    @property
    def pos_thresmaxx(self):
        """Maximum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x position threshold for each BPM.
        """
        return _np.array([b.pos_thresmaxx for b in self._devices])

    @pos_thresmaxx.setter
    def pos_thresmaxx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.pos_thresmaxx = value[idx]

    @property
    def pos_thresminy(self):
        """Minimum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y position threshold for each BPM.
        """
        return _np.array([b.pos_thresminy for b in self._devices])

    @pos_thresminy.setter
    def pos_thresminy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.pos_thresminy = value[idx]

    @property
    def pos_thresmaxy(self):
        """Maximum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y position threshold for each BPM.
        """
        return _np.array([b.pos_thresmaxy for b in self._devices])

    @pos_thresmaxy.setter
    def pos_thresmaxy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.pos_thresmaxy = value[idx]

    @property
    def pos_inst_lower(self):
        """Instantaneous position interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_lower for b in self._devices])

    @property
    def pos_inst_upper(self):
        """Instantaneous position interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_upper for b in self._devices])

    @property
    def pos_inst_lower_x(self):
        """Instantaneous position interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_lower_x for b in self._devices])

    @property
    def pos_inst_upper_x(self):
        """Instantaneous position interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_upper_x for b in self._devices])

    @property
    def pos_inst_lower_y(self):
        """Instantaneous position interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_lower_y for b in self._devices])

    @property
    def pos_inst_upper_y(self):
        """Instantaneous position interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_upper_y for b in self._devices])

    @property
    def pos_latch_lower(self):
        """Latch position interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_lower for b in self._devices])

    @property
    def pos_latch_upper(self):
        """Latch position interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_upper for b in self._devices])

    @property
    def pos_latch_lower_x(self):
        """Latch position interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_lower_x for b in self._devices])

    @property
    def pos_latch_upper_x(self):
        """Latch position interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_upper_x for b in self._devices])

    @property
    def pos_latch_lower_y(self):
        """Latch position interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_lower_y for b in self._devices])

    @property
    def pos_latch_upper_y(self):
        """Latch position interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_upper_y for b in self._devices])

    #  --- angulation interlock ---

    def cmd_ang_enable(self, timeout=TIMEOUT):
        """Enable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkAngEn-Sts', 1, timeout=timeout)

    def cmd_ang_disable(self, timeout=TIMEOUT):
        """Disable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkAngEn-Sts', 0, timeout=timeout)

    def cmd_reset_ang(self):
        """Reset all BPM angulation interlock."""
        for dev in self.devices:
            dev.cmd_reset_ang()
        return True

    @property
    def ang_thresminx(self):
        """Minimum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresminx for b in self._devices])

    @ang_thresminx.setter
    def ang_thresminx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresminx = value[idx]

    @property
    def ang_thresmaxx(self):
        """Maximum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresmaxx for b in self._devices])

    @ang_thresmaxx.setter
    def ang_thresmaxx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresmaxx = value[idx]

    @property
    def ang_thresminy(self):
        """Minimum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresminy for b in self._devices])

    @ang_thresminy.setter
    def ang_thresminy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresminy = value[idx]

    @property
    def ang_thresmaxy(self):
        """Maximum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresmaxy for b in self._devices])

    @ang_thresmaxy.setter
    def ang_thresmaxy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresmaxy = value[idx]

    @property
    def ang_inst_lower(self):
        """Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_lower for b in self._devices])

    @property
    def ang_inst_upper(self):
        """Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_upper for b in self._devices])

    @property
    def ang_inst_lower_x(self):
        """Instantaneous angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_lower_x for b in self._devices])

    @property
    def ang_inst_upper_x(self):
        """Instantaneous angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_upper_x for b in self._devices])

    @property
    def ang_inst_lower_y(self):
        """Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_lower_y for b in self._devices])

    @property
    def ang_inst_upper_y(self):
        """Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_upper_y for b in self._devices])

    @property
    def ang_latch_lower(self):
        """Latch angulation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_lower for b in self._devices])

    @property
    def ang_latch_upper(self):
        """Latch angulation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_upper for b in self._devices])

    @property
    def ang_latch_lower_x(self):
        """Latch angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_lower_x for b in self._devices])

    @property
    def ang_latch_upper_x(self):
        """Latch angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_upper_x for b in self._devices])

    @property
    def ang_latch_lower_y(self):
        """Latch angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_lower_y for b in self._devices])

    @property
    def ang_latch_upper_y(self):
        """Latch angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_upper_y for b in self._devices])

    @property
    def slow_orbit(self):
        """Slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self._devices:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    @property
    def possum(self):
        """Sum vector, at Monit rate.

        Returns:
            possum (numpy.ndarray, 160): Sum vector, at Monit rate.
        """
        return _np.array([b.possum for b in self._devices])

    @property
    def position(self):
        """Position vectors.

        Position at each BPM is defined as:
            (pos BPM downstream + pos BPM upstream)/2

        Returns:
            posx (numpy.ndarray, 160): Horizontal Position.
            posy (numpy.ndarray, 160): Vertical Position.
        """
        orbx, orby = self.slow_orbit
        posx = _np.array(self.calc_intlk_metric(orbx, metric='pos'))
        posy = _np.array(self.calc_intlk_metric(orby, metric='pos'))
        return posx, posy

    @property
    def angulation(self):
        """Angulation vectors.

        Angulation at each BPM is defined as:
            (posição BPM downstream - posição BPM upstream)

        Returns:
            angx (numpy.ndarray, 160): Horizontal Angulation.
            angy (numpy.ndarray, 160): Vertical Angulation.
        """
        orbx, orby = self.slow_orbit
        angx = _np.array(self.calc_intlk_metric(orbx, metric='ang'))
        angy = _np.array(self.calc_intlk_metric(orby, metric='ang'))
        return angx, angy

    def _handle_thres_input(self, value):
        if isinstance(value, (int, float, _np.int_, _np.float_)):
            return [value, ] * len(self.BPM_NAMES)
        if len(value) != len(self.devices):
            raise ValueError(
                'the expected length is {0}, not {1}'.format(
                    len(self.devices), len(value)))
        return value



#!/usr/local/bin/python-sirius
"""Driver module."""

from pcaspy import Alarm as _Alarm, Severity as _Severity

from ...namesys import SiriusPVName
from ...pwrsupply.csdev import get_ps_interlocks as _get_ps_interlocks
from ..app import App as _App
from ..pvs import ComputedPV as _ComputedPV
from .pvs import PSStatusPV as _PSStatusPV, PSDiffPV as _PSDiffPV


class PSDiagApp(_App):
    """Main application responsible for updating DB."""

    def _create_computed_pvs(self, psnames):
        self._psnames = psnames
        for psname in self._psnames:
            devname = SiriusPVName(psname).substitute(prefix=self._prefix)

            # DiagCurrentDiff-Mon
            nrpvs = 4 if devname.sec != 'LI' else 2
            pvs = [None]*nrpvs
            pvs[_PSDiffPV.CURRT_SP] = devname + ':Current-SP'
            pvs[_PSDiffPV.CURRT_MON] = devname + ':Current-Mon'
            if devname.sec != 'LI':
                pvs[_PSDiffPV.CURRT_REF] = devname + ':CurrentRef-Mon'
                pvs[_PSDiffPV.OPMODESTS] = devname + ':OpMode-Sts'
            pvo = _ComputedPV(
                psname + ':DiagCurrentDiff-Mon', _PSDiffPV(), self._queue,
                pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagStatus-Mon
            computer = _PSStatusPV()
            if devname.sec != 'LI':
                intlks = _get_ps_interlocks(psname=psname)
                intlk_list = [':' + ppt for ppt in intlks if 'Intlk' in ppt]
                alarm_list = [':' + ppt for ppt in intlks if 'Alarm' in ppt]

                if psname in ['BO-Fam:PS-B-1', 'BO-Fam:PS-B-2']:
                    for aux in ['a', 'b', 'c']:
                        intlk_list.extend(
                            [aux+':'+ilk for ilk in intlks if 'Intlk' in ilk
                             if 'Soft' not in ilk and 'Hard' not in ilk])
                        alarm_list.extend(
                            [aux+':'+alm for alm in intlks if 'Alarm' in alm])

                nbpvs = 5
                pvs = [None]*(nbpvs+len(intlk_list)+len(alarm_list))
                pvs[_PSStatusPV.PWRSTE_STS] = devname + ':PwrState-Sts'
                pvs[_PSStatusPV.CURRT_DIFF] = devname + ':DiagCurrentDiff-Mon'
                pvs[_PSStatusPV.OPMODE_SEL] = devname + ':OpMode-Sel'
                pvs[_PSStatusPV.OPMODE_STS] = devname + ':OpMode-Sts'
                if psname.dev not in ['FCH', 'FCV']:
                    pvs[_PSStatusPV.WAVFRM_MON] = devname + ':Wfm-Mon'
                else:
                    pvs[_PSStatusPV.TRIGEN_STS] = devname + ':TrigEn-Sts'

                computer.INTLK_PVS = list()
                for idx, intlk in enumerate(intlk_list):
                    pvidx = idx + nbpvs
                    computer.INTLK_PVS.append(pvidx)
                    pvs[pvidx] = devname + intlk
                computer.ALARM_PVS = list()
                for idx, alarm in enumerate(alarm_list):
                    pvidx = idx + nbpvs + len(intlk_list)
                    computer.ALARM_PVS.append(pvidx)
                    pvs[pvidx] = devname + alarm
            else:
                pvs = [None]*5
                pvs[_PSStatusPV.PWRSTE_STS] = devname + ':PwrState-Sts'
                pvs[_PSStatusPV.CURRT_DIFF] = devname + ':DiagCurrentDiff-Mon'
                pvs[_PSStatusPV.INTRLCK_LI] = devname + ':StatusIntlk-Mon'
                pvs[_PSStatusPV.WARNSTS_LI] = devname + ':IntlkWarn-Mon'
                pvs[_PSStatusPV.CONNCTD_LI] = devname + ':Connected-Mon'

            pvo = _ComputedPV(
                psname + ':DiagStatus-Mon', computer, self._queue,
                pvs, monitor=False)
            self.pvs.append(pvo)
        self._pvs_connected = {pvo: False for pvo in self.pvs}

    def _update_pvs(self):
        for pvo in self.pvs:
            if not pvo.connected:
                if self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.TIMEOUT_ALARM,
                        severity=_Severity.INVALID_ALARM,
                        field='status')
                self._pvs_connected[pvo] = False
                if 'DiagStatus' in pvo.pvname:
                    self.run_callbacks(pvo.pvname, value=pvo.value)
            else:
                if not self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.NO_ALARM,
                        severity=_Severity.NO_ALARM, field='status')
                self._pvs_connected[pvo] = True
                self.run_callbacks(pvo.pvname, value=pvo.value)



#!/usr/bin/env python-sirius

"""Unittest module for opticscorr.py."""

from unittest import TestCase
from siriuspy import util
from siriuspy.opticscorr import csdev
from siriuspy.opticscorr.csdev import \
    get_chrom_database, \
    get_tune_database


PUB_INTERFACE = (
        'ETypes',
        'Const',
        'get_chrom_database',
        'get_tune_database',
    )


class TestOpticsCorrCSDevice(TestCase):
    """Test siriuspy.opticscorr.csdev module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_get_chrom_database(self):
        """Test get_chrom_database."""
        for accelerator in ('SI', 'BO'):
            if accelerator == 'SI':
                sfams = ('SFA1', 'SFA2', 'SDA1', 'SDA2', 'SDA3',
                         'SFB1', 'SFB2', 'SDB1', 'SDB2', 'SDB3',
                         'SFP1', 'SFP2', 'SDP1', 'SDP2', 'SDP3')
            elif accelerator == 'BO':
                sfams = ('SF', 'SD')
            dbase = get_chrom_database(accelerator)
            self.assertIsInstance(dbase, dict)
            self.assertTrue('Version-Cte' in dbase)
            self.assertTrue('Log-Mon' in dbase)
            self.assertTrue('ChromX-SP' in dbase)
            self.assertTrue('ChromX-RB' in dbase)
            self.assertTrue('ChromY-SP' in dbase)
            self.assertTrue('ChromY-RB' in dbase)
            self.assertTrue('DeltaChromX-SP' in dbase)
            self.assertTrue('DeltaChromX-RB' in dbase)
            self.assertTrue('DeltaChromY-SP' in dbase)
            self.assertTrue('DeltaChromY-RB' in dbase)
            self.assertTrue('ChromY-Mon' in dbase)
            self.assertTrue('ChromY-Mon' in dbase)
            self.assertTrue('CalcChromX-Mon' in dbase)
            self.assertTrue('CalcChromY-Mon' in dbase)
            self.assertTrue('ApplyDelta-Cmd' in dbase)
            self.assertTrue('ConfigName-SP' in dbase)
            self.assertTrue('ConfigName-RB' in dbase)
            self.assertTrue('RespMat-Mon' in dbase)
            self.assertTrue('NominalChrom-Mon' in dbase)
            self.assertTrue('NominalSL-Mon' in dbase)
            self.assertTrue('ConfigPS-Cmd' in dbase)
            self.assertTrue('Status-Mon' in dbase)
            self.assertTrue('StatusLabels-Cte' in dbase)

            for fam in sfams:
                self.assertTrue('SL' + fam + '-Mon' in dbase)
                self.assertEqual(dbase['SL' + fam + '-Mon']['unit'], '1/m^2')
            if accelerator == 'SI':
                self.assertTrue('CorrMeth-Sel' in dbase)
                self.assertTrue('CorrMeth-Sts' in dbase)
                # self.assertTrue('SyncCorr-Sel' in dbase)
                # self.assertTrue('SyncCorr-Sts' in dbase)
                # self.assertTrue('ConfigTiming-Cmd' in dbase)

    def test_get_tune_database(self):
        """Test get_tune_database."""
        for accelerator in ('SI', 'BO'):
            if accelerator == 'SI':
                qfams = ('QFA', 'QFB', 'QFP',
                         'QDA', 'QDB1', 'QDB2', 'QDP1', 'QDP2')
            elif accelerator == 'BO':
                qfams = ('QF', 'QD')
            dbase = get_tune_database(accelerator)
            self.assertIsInstance(dbase, dict)
            self.assertTrue('Version-Cte' in dbase)
            self.assertTrue('Log-Mon' in dbase)
            self.assertTrue('DeltaTuneX-SP' in dbase)
            self.assertTrue('DeltaTuneX-RB' in dbase)
            self.assertTrue('DeltaTuneY-SP' in dbase)
            self.assertTrue('DeltaTuneY-RB' in dbase)
            self.assertTrue('ApplyDelta-Cmd' in dbase)
            self.assertTrue('RespMat-Mon' in dbase)
            self.assertTrue('NominalKL-Mon' in dbase)
            self.assertTrue('ConfigPS-Cmd' in dbase)
            self.assertTrue('Status-Mon' in dbase)
            self.assertTrue('StatusLabels-Cte' in dbase)

            for fam in qfams:
                self.assertTrue('RefKL'+fam+'-Mon' in dbase)
                self.assertEqual(dbase['RefKL'+fam+'-Mon']['unit'], '1/m')
                self.assertTrue('DeltaKL'+fam+'-Mon' in dbase)
                self.assertEqual(dbase['DeltaKL'+fam+'-Mon']['unit'], '1/m')
            if accelerator == 'SI':
                self.assertTrue('CorrMeth-Sel' in dbase)
                self.assertTrue('CorrMeth-Sts' in dbase)
                self.assertTrue('SyncCorr-Sel' in dbase)
                self.assertTrue('SyncCorr-Sts' in dbase)
                self.assertTrue('ConfigTiming-Cmd' in dbase)



"""PRU Controller Parameters.

These parameters are used in PRUController class to
interact with BSMP devices.
"""

from ..bsmp.constants import ConstPSBSMP as _const
from ..psctrl.psmodel import PSModelFactory as _PSModelFactory


class PRUCParmsFBP:
    """FBP-specific PRUC parameters."""

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FBP')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))
    groups[CONST_PSBSMP.G_SOFB] = (
        CONST_PSBSMP.V_PS_SETPOINT1,
        CONST_PSBSMP.V_PS_SETPOINT2,
        CONST_PSBSMP.V_PS_SETPOINT3,
        CONST_PSBSMP.V_PS_SETPOINT4,
        CONST_PSBSMP.V_PS_REFERENCE1,
        CONST_PSBSMP.V_PS_REFERENCE2,
        CONST_PSBSMP.V_PS_REFERENCE3,
        CONST_PSBSMP.V_PS_REFERENCE4,
        CONST_PSBSMP.V_I_LOAD1,
        CONST_PSBSMP.V_I_LOAD2,
        CONST_PSBSMP.V_I_LOAD3,
        CONST_PSBSMP.V_I_LOAD4,)


class PRUCParmsFBP_DCLink:
    """FBP_DCLink-specific PRUC parameters."""

    FREQ_SCAN = 2.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FBP_DCLink')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAC_2S_DCDC:
    """FAC_2S specific PRUC parameters.

    Represent FAC_2S_DCDC psmodels.
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAC_2S_DCDC')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAC_2P4S_DCDC:
    """FAC-specific PRUC parameters.

    Represent FAC_2P4S psmodels.
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAC_2P4S_DCDC')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAC_DCDC:
    """FAC-specific PRUC parameters.

    Represent FAC, FAC_2S, FAC_2P4S psmodels.
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAC_DCDC')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAC_2S_ACDC:
    """FAC_2S_ACDC-specific PRUC parameters."""

    FREQ_SCAN = 2.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAC_2S_ACDC')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAC_2P4S_ACDC:
    """FAC_2P4S_ACDC-specific PRUC parameters."""

    FREQ_SCAN = 2.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAC_2P4S_ACDC')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAP:
    """FAC-specific PRUC parameters.

    Represent FAP
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAP')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAP_4P:
    """FAC-specific PRUC parameters.

    Represent FAP_4P
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAP_4P')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))


class PRUCParmsFAP_2P2S:
    """FAC_2P2S-specific PRUC parameters.

    Represent FAP_2P2S
    """

    FREQ_SCAN = 5.0  # [Hz]

    # PS model parms
    model = _PSModelFactory.create('FAP_2P2S')
    CONST_PSBSMP = model.bsmp_constants
    Entities = model.entities

    groups = dict()
    # reserved variable groups (not to be used)
    groups[_const.G_ALL] = tuple(sorted(Entities.list_variables(0)))
    groups[_const.G_READONLY] = tuple(sorted(Entities.list_variables(1)))
    groups[_const.G_WRITE] = tuple(sorted(Entities.list_variables(2)))



"""Define properties of all timing devices and their connections."""

import ast as _ast
from copy import deepcopy as _dcopy
from threading import Lock as _Lock
from siriuspy import clientweb as _web
from siriuspy.namesys import SiriusPVName as _PVName, Filter as _Filter
from .ll_time_search import LLTimeSearch as _LLSearch

_timeout = 1.0


class HLTimeSearch:
    """Contain properties of the triggers."""

    _hl_triggers = dict()
    _hl_events = dict()
    COMMOM_EVTS = ('Dsbld', 'RmpBO', 'Linac', 'Study')

    _lock = _Lock()

    @classmethod
    def get_hl_events(cls):
        """Return a dictionary with high level events."""
        cls._init()
        return _dcopy(cls._hl_events)

    @classmethod
    def get_configurable_hl_events(cls):
        """Return a dictionary with configurable high level events."""
        cls._init()
        return {e: c for e, c in cls._hl_events.items() if 0 < int(c[3:]) < 64}

    @classmethod
    def get_hl_triggers(cls, filters=None, sorting=None):
        """Return a dictionary with high level triggers."""
        cls._init()
        all_devs = sorted(cls._hl_triggers.keys())
        return _Filter.process_filters(
            all_devs, filters=filters, sorting=sorting)

    @classmethod
    def get_hl_trigger_predef_db(cls, hl_trigger):
        """Return the default database of the high level trigger."""
        cls._init()
        dic_ = _dcopy(cls._hl_triggers[hl_trigger]['database'])
        src = dic_['Src']
        src['enums'] = cls.COMMOM_EVTS + src['enums']
        return dic_

    @classmethod
    def get_hl_trigger_allowed_evts(cls, hl_trigger):
        """Return the possible events of the high level trigger."""
        cls._init()
        enums = cls._hl_triggers[hl_trigger]['database']['Src']['enums']
        return _dcopy(enums)

    @classmethod
    def get_hl_trigger_prop_value(cls, hl_trigger, prop='Delay'):
        """Return the default value of the property prop of the HL trigger."""
        cls._init()
        val = cls._hl_triggers[hl_trigger]['database'][prop]['value']
        if prop == 'Src':
            val += 1 if cls.has_clock(hl_trigger) else 0
        return val

    @classmethod
    def get_hl_trigger_prop_limits(cls, hl_trigger, prop='Delay'):
        """Return the default limits of the property prop of the HL trigger."""
        cls._init()
        dic_ = dict()
        for lim in {'lolo', 'hihi', 'high', 'low', 'lolim', 'hilim'}:
            val = cls._hl_triggers[hl_trigger]['database'][prop].get(lim)
            if val is not None:
                dic_[lim] = val
        return dic_

    @classmethod
    def get_ll_trigger_names(cls, hl_trigger):
        """Get Low Level trigger object names."""
        cls._init()
        return sorted(cls._hl2ll_trigs.get(hl_trigger, list()))

    @classmethod
    def get_hl_from_ll_triggers(cls, channel):
        # channel = _LLSearch.get_channel_output_port_pvname(channel)
        prpt = channel.propty
        if prpt.startswith('OTP'):
            val = int(prpt[3:]) - 12
            newprpt = 'OUT'+str(val) if 0 <= val < 8 else 'OTP'+str(val+12)
            channel = channel.substitute(propty=newprpt)
        return cls._ll2hl_trigs.get(channel, '')

    @classmethod
    def get_hl_trigger_channels(cls, hl_trigger):
        cls._init()
        ret = cls._hl_triggers.get(hl_trigger)
        chans = ret['channels'] if ret else list()
        return sorted(chans)

    @classmethod
    def has_delay_type(cls, hl_trigger):
        """Return True if hl_trigger has property delayType."""
        cls._init()
        ll_chans = cls.get_ll_trigger_names(hl_trigger)
        return all([_LLSearch.has_delay_type(name) for name in ll_chans])

    @classmethod
    def has_clock(cls, hl_trigger):
        """Return True if hl_trigger can listen to Clocks from EVG."""
        cls._init()
        ll_chans = cls.get_ll_trigger_names(hl_trigger)
        return all([_LLSearch.has_clock(name) for name in ll_chans])

    @classmethod
    def check_hl_triggers_consistency(cls):
        """Check consitency of Triggers definition.

        Check if High Level definition of Triggers is consistent with
        Low Level connections of the timing devices.
        """
        cls._init()
        twds_evg = _LLSearch.get_connections_twds_evg()
        from_evg = _LLSearch.get_connections_from_evg()
        for trig, val in cls._hl_triggers.items():
            chans = {_PVName(chan) for chan in val['channels']}
            for chan in chans:
                tmp = twds_evg.get(chan)
                if tmp is None:
                    raise Exception(
                        chan + ' is used in HL trigger ' + trig +
                        ' but not specified in timing connections data.')
                if not tmp:
                    raise Exception(
                        chan + ' which is used in HL trigger ' + trig +
                        ' maybe were already used somewhere else.')
                up_dev = tmp.pop()
                diff_devs = from_evg[up_dev] - chans
                if diff_devs and not chan.dev.endswith('BPM'):
                    raise Exception(
                        'Devices: ' + ' '.join(diff_devs) +
                        ' are connected to the same output of ' +
                        up_dev + ' as ' + chan +
                        ' but are not related to the same trigger ('+trig+').')

    @classmethod
    def reset(cls):
        cls._hl_triggers = dict()
        cls._init()

    @classmethod
    def _init(cls):
        """Initialize the Instance."""
        with cls._lock:
            if cls._hl_triggers:
                return
            if not _web.server_online():
                raise Exception('Could not connect with Consts Server!!')
            text1 = _web.high_level_triggers(timeout=_timeout)
            text2 = _web.high_level_events(timeout=_timeout)
            temp_dict = _ast.literal_eval(text1)
            hl_trigs = dict()
            for k, vs in temp_dict.items():
                vs['channels'] = tuple(map(_PVName, vs['channels']))
                hl_trigs[_PVName(k)] = vs
            cls._hl_triggers = hl_trigs
            cls._hl_events = _ast.literal_eval(text2)

            hl2ll = dict()
            for hlt, val in cls._hl_triggers.items():
                chans = val['channels']
                hl2ll[hlt] = {_LLSearch.get_trigger_name(c) for c in chans}
            ll2hl = dict()
            for k, vs in hl2ll.items():
                ll2hl.update({v: k for v in vs})
            cls._hl2ll_trigs = hl2ll
            cls._ll2hl_trigs = ll2hl



#!/usr/bin/env python-sirius

"""Test serial module."""

from unittest import TestCase
from unittest.mock import Mock

from siriuspy.bsmp import (
    Channel,
    Message,
    Package,
    SerialErrMsgShort,
    SerialError,
    SerialErrPckgLen,
)
from siriuspy.util import check_public_interface_namespace


class TestBSMPMessage(TestCase):
    """Test BSMP Message class."""

    api = (
        'message',
        'stream',
        'cmd',
        'size',
        'payload',
    )

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(Message, self.api))

    def test_init(self):
        """Test constructor that creates object from stream."""
        stream = ['\x01', '\x00', '\x03', '\x02', '\n', '\x00']
        m = Message(stream)
        # Check properties
        self.assertEqual(m.cmd, 0x01)
        self.assertEqual(m.payload, [chr(2), chr(10), chr(0)])
        # Convert message to stream
        self.assertEqual(m.stream, stream)

    def test_small_message(self):
        """Test message with stream impossibly small."""
        with self.assertRaises(SerialErrMsgShort):
            Message(['\x11', '\x00'])

    def test_message_with_no_load(self):
        """Test constructor with no load."""
        m = Message.message(cmd=0x00)
        self.assertEqual(m.cmd, 0x00)
        self.assertEqual(m.payload, [])

    def test_message_with_load(self):
        """Test constructor with load."""
        m = Message.message(cmd=0x10, payload=[1])
        self.assertEqual(m.cmd, 0x10)
        self.assertEqual(m.payload, [1])

    def test_message_with_extraneous_load(self):
        """Test constructor with loads that are not list."""
        loads = [1, 'string', (1, 2, 3), 63.7]
        for load in loads:
            with self.assertRaises(TypeError):
                Message.message(cmd=0x10, payload=load)

    def test_message_with_big_load(self):
        """Test constructor with load bigger than 65535."""
        load = [0 for _ in range(65536)]
        with self.assertRaises(ValueError):
            Message.message(cmd=0x10, payload=load)

    def test_conv_to_stream(self):
        """Test conversion from message to stream."""
        # Example in 3.8.2 of BSMP protocol document
        curve_id = chr(0x07)
        blk_n = [chr(0x40), chr(0x00)]
        blk_load = [chr(0xDD) for _ in range(16384)]
        load = [curve_id] + blk_n + blk_load
        m = Message.message(cmd=0x41, payload=load)

        expected_stream = [chr(0x41), chr(0x40), chr(0x03)] + load

        self.assertEqual(m.stream, expected_stream)


class TestBSMPPackage(TestCase):
    """Test BSMP Package class."""

    # Tuples with address, message and checksum
    data = [
        (1, 0x10, [chr(3)], ['\x01', '\x10', '\x00', '\x01', '\x03', chr(235)], 235),
        (0, 0x11, [chr(3), chr(255), chr(255)], ['\x00', '\x11', '\x00', '\x03', '\x03', '\xFF', '\xFF', chr(235)], 235),
        (2, 0x20, [chr(4), chr(1), chr(187), chr(187)], ['\x02', '\x20', '\x00', '\x04', '\x04', '\x01', '\xBB', '\xBB', chr(95)], 95),
        (3, 0x22,
            [chr(2), chr(1), chr(187), chr(187), chr(1), chr(187), chr(187), chr(1), chr(187), chr(187), chr(1), chr(187), chr(187), chr(204)],
            ['\x03', '\x22', '\x00', '\x0E', '\x02', '\x01', '\xBB', '\xBB', '\x01', '\xBB', '\xBB', '\x01', '\xBB', '\xBB', '\x01', '\xBB', '\xBB', '\xCC', chr(35)],
            35,
        ),
    ]

    api = (
        'package',
        'address',
        'message',
        'checksum',
        'stream',
        'calc_checksum',
        'verify_checksum',
    )

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(Package, self.api))

    def test_package(self):
        """Test constructor."""
        address = 1
        m = Message.message(0x00)
        p = Package.package(address=address, message=m)

        self.assertEqual(p.address, address)
        self.assertEqual(p.message, m)

    def test_package_stream(self):
        """Test constructor that creates object from stream."""
        for d in self.data:
            stream = d[3]
            p = Package(stream)
            self.assertEqual(p.address, d[0])
            self.assertEqual(p.message.cmd, d[1])
            self.assertEqual(p.message.payload, d[2])

    def test_parse_small_stream(self):
        """Test constructor that tries to parse strem smaller than 5."""
        stream = ['\x02', '\x00', '\x00', chr(254)]
        with self.assertRaises(SerialErrPckgLen):
            Package(stream)

    def test_checksum(self):
        """Test checksum value."""
        for d in self.data:
            p = Package.package(d[0], Message.message(d[1], payload=d[2]))
            self.assertEqual(p.checksum, d[4])

    def test_conv_to_stream(self):
        """Test conversion of package to stream."""
        for d in self.data:
            p = Package.package(d[0], Message.message(d[1], payload=d[2]))
            self.assertEqual(p.stream, d[3])

    def test_verify_checksum(self):
        """Verify checksum sucessfully."""
        for d in self.data:
            stream = d[3]
            self.assertTrue(Package.verify_checksum(stream))

    def test_verify_false_checksum(self):
        """Verify checksums fail."""
        for d in self.data:
            stream, checksum = d[3][:-1], ord(d[3][-1])
            stream += [chr(checksum + 1)]
            self.assertFalse(Package.verify_checksum(stream))


class TestBSMPChannel(TestCase):
    """Test Channel class of BSMP package."""

    api = (
        'LOCK',
        'iointerf',
        'address',
        'size_counter',
        'size_counter_reset',
        'read',
        'write',
        'request_',
        'request',
        'create_lock',
    )

    def setUp(self):
        """Setup common to all tests."""
        self.serial = Mock()
        self.channel = Channel(self.serial, 1)

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(Channel, self.api))

    def test_read_calls_serial_method(self):
        """Test UART_read is called."""
        response = Message.message(0x11, payload=[chr(10)])
        self.serial.UART_read.return_value = Package.package(0x01, response).stream
        self.channel.read()
        self.serial.UART_read.assert_called_once()

    def test_read(self):
        """Test read method."""
        response = Message.message(0x11, payload=[chr(10)])
        self.serial.UART_read.return_value = Package.package(0x01, response).stream
        recv = self.channel.read()
        self.assertEqual(recv.cmd, response.cmd)
        self.assertEqual(recv.payload, response.payload)

    def test_write_calls_serial_method(self):
        """Test write calls UART_write."""
        message = Message.message(0x10, payload=[chr(1)])
        expected_stream = Package.package(self.channel.address, message).stream
        self.channel.write(message, 1000)
        self.serial.UART_write.assert_called_with(expected_stream, timeout=1000)

    def test_request(self):
        """Test request."""
        response = Message.message(0x11, payload=[chr(10)])
        self.serial.UART_request.return_value = Package.package(0x01, response).stream
        recv = self.channel.request(Message.message(0x01, payload=[chr(1)]), timeout=1)
        self.assertEqual(recv.cmd, response.cmd)
        self.assertEqual(recv.payload, response.payload)

    def test_request_fail(self):
        """Test exception is raised when serial fails."""
        self.serial.UART_request.return_value = None
        with self.assertRaises(SerialError):
            self.channel.request(Message.message(0x10, payload=[chr(10)]))



"""SiriusPVTimeSerie Class."""

import collections as _collections
import time as _time
import threading as _threading
import numpy as _np


class SiriusPVTimeSerie:
    """Class to handle time series from pv monitoring."""

    def __init__(self, pv,
                 time_window=None,
                 nr_max_points=None,
                 time_min_interval=0.0,
                 mode=0,
                 timestamp_init_data=None,
                 value_init_data=None,
                 use_pv_timestamp=True):
        """Class constructor."""
        if (use_pv_timestamp is False) and (mode == 1):
            raise ValueError(
                "Can not create an auto-fill serie without using "
                "PV timestamp!")
        self._pvobj = pv
        self._time_window = time_window
        self._time_min_interval = time_min_interval
        self._nr_max_points = nr_max_points
        self._use_pv_timestamp = use_pv_timestamp
        if timestamp_init_data:
            if not value_init_data:
                raise ValueError("Provide 'value_init_data' input!")
            self._timestamp_deque = _collections.deque(
                timestamp_init_data, maxlen=nr_max_points)
            self._value_deque = _collections.deque(
                value_init_data, maxlen=nr_max_points)
        else:
            self._timestamp_deque = _collections.deque(maxlen=nr_max_points)
            self._value_deque = _collections.deque(maxlen=nr_max_points)
        self._mode = mode
        if self._mode == 1:
            self._th_auto_acquire = _threading.Thread(
                target=self._auto_acquire, daemon=True)
            self._stop_auto_acquire = False
            self._th_auto_acquire.start()
        else:
            self._stop_auto_acquire = True

    @property
    def time_window(self):
        """Time window of datapoints."""
        return self._time_window

    @time_window.setter
    def time_window(self, value):
        self._time_window = value
        timestamp = _time.time()
        self._update(timestamp)

    @property
    def time_min_interval(self):
        """Minimum time interval between the datapoints."""
        return self._time_min_interval

    @time_min_interval.setter
    def time_min_interval(self, value):
        self._time_min_interval = value

        if self._timestamp_deque:
            old_timestamp_deque = _collections.deque(
                self._timestamp_deque, maxlen=self._nr_max_points)
            old_value_deque = _collections.deque(
                self._value_deque, maxlen=self._nr_max_points)
            self._timestamp_deque = _collections.deque(
                maxlen=self._nr_max_points)
            self._value_deque = _collections.deque(
                maxlen=self._nr_max_points)

            aux = old_timestamp_deque.pop()
            self._timestamp_deque.appendleft(aux)
            aux = old_value_deque.pop()
            self._value_deque.appendleft(aux)

            while old_timestamp_deque:
                if self._time_min_interval < (
                        self._timestamp_deque[0] - old_timestamp_deque[-1]):
                    aux = old_timestamp_deque.pop()
                    self._timestamp_deque.appendleft(aux)
                    aux = old_value_deque.pop()
                    self._value_deque.appendleft(aux)
                else:
                    old_timestamp_deque.pop()
                    old_value_deque.pop()

    @property
    def nr_max_points(self):
        """Maximum number of datapoints."""
        return self._nr_max_points

    @nr_max_points.setter
    def nr_max_points(self, value):
        self._nr_max_points = value
        self._timestamp_deque = _collections.deque(
            self._timestamp_deque, maxlen=self._nr_max_points)
        self._value_deque = _collections.deque(
            self._value_deque, maxlen=self._nr_max_points)

    @property
    def mode(self):
        """Mode of acquisition: 0 if manual and 1 if automatic."""
        return self._mode

    @mode.setter
    def mode(self, value):
        if self._mode != value:
            self._mode = value
            if self._mode == 1:
                self._th_auto_acquire = _threading.Thread(
                    target=self._auto_acquire, daemon=True)
                self._stop_auto_acquire = False
                self._th_auto_acquire.start()
            else:
                self._stop_auto_acquire = True

    @property
    def serie(self):
        """PV time series, as two separate lists: timestamp and value."""
        return self.get_serie()

    def get_serie(self, time_absolute=False):
        """Return series, as two separate numpy arrays: timestamp and value."""
        timestamp = _time.time()
        self._update(timestamp)
        timestamp_array = _np.array(self._timestamp_deque)
        if not time_absolute:
            timestamp_array -= timestamp
        value_array = _np.array(self._value_deque)
        return timestamp_array, value_array

    def acquire(self):
        """Acquire a new time serie datapoint.

        Returns True if datapoint was acquired and False otherwise.
        """
        minintv = self._time_min_interval
        # check if pv is connected
        if self.connected():
            timestamp = _time.time()
            if self._use_pv_timestamp:
                pv_timestamp, pv_value = \
                    self._pvobj.timestamp, self._pvobj.value
            else:
                pv_timestamp, pv_value = timestamp, self._pvobj.value

            # check if it is a new datapoint
            if not self._timestamp_deque or \
                    pv_timestamp != self._timestamp_deque[-1]:
                # check if there is a limiting time_window
                if self._time_window is None:
                    # check if there is a limiting time_min_interval
                    if not self._timestamp_deque or \
                            minintv <= timestamp-self._timestamp_deque[-1]:
                        self._timestamp_deque.append(pv_timestamp)
                        self._value_deque.append(pv_value)
                        return True
                    else:
                        # print('not acquired: time interval not sufficient')
                        return False
                else:
                    # Check if the datapoints in the deques are yet valid
                    # to the limiting time_window
                    self._update(timestamp)
                    # check if the new point is within the limiting time_window
                    if pv_timestamp >= timestamp - self._time_window:
                        if not self._timestamp_deque or \
                                minintv <= timestamp-self._timestamp_deque[-1]:
                            self._timestamp_deque.append(pv_timestamp)
                            self._value_deque.append(pv_value)
                            return True
                        else:
                            # print('not acquired: not enough time interval')
                            return False
                    else:
                        # print('not acquired: not within time_window')
                        return False
            else:
                # print('not acquired: item already in deque')
                return False
        else:
            # print('not acquired: pv not connected')
            return False

    def _auto_acquire(self):
        while not self._stop_auto_acquire:
            self.acquire()
            _time.sleep(self.time_min_interval)

    def _update(self, timestamp):
        """Update time serie according to current timestamp."""
        if not self._timestamp_deque or self._time_window is None:
            return

        min_timestamp = timestamp - self._time_window
        if self._timestamp_deque[-1] <= min_timestamp:
            self.clearserie()
        elif self._timestamp_deque[0] >= min_timestamp:
            pass
        else:
            low_interval_end = 0
            high_interval_end = len(self._timestamp_deque)-1
            search_index = (high_interval_end - low_interval_end)//2

            while low_interval_end != search_index:
                if self._timestamp_deque[search_index] <= min_timestamp:
                    low_interval_end = search_index
                    search_index = (
                        (high_interval_end - low_interval_end)//2 +
                        low_interval_end)
                else:
                    high_interval_end = search_index
                    search_index = (
                        (high_interval_end - low_interval_end)//2 +
                        low_interval_end)

            for _ in range(search_index + 1):
                self._timestamp_deque.popleft()
                self._value_deque.popleft()

    def clearserie(self):
        """Clear time serie."""
        self._timestamp_deque.clear()
        self._value_deque.clear()

    def connected(self):
        """Check PV connection."""
        return self._pvobj.connected

    def __str__(self):
        """Return string representation of time series."""
        raise NotImplementedError()



"""Unittest module for ma_search.py."""

from unittest import TestCase

from siriuspy import util
from siriuspy.search import ma_search
from siriuspy.search import MASearch


class TestModule(TestCase):
    """Test Search module."""

    public_interface = (
        'MASearch',
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                                ma_search, TestModule.public_interface)
        self.assertTrue(valid)


class TestMASearch(TestCase):
    """Test MASearch."""

    public_interface = (
        'get_manames',
        'get_pwrsupply_manames',
        'get_mapositions',
        'get_manicknames',
        'conv_maname_2_trims',
        'conv_maname_2_magfunc',
        'conv_maname_2_psnames',
        'conv_psmaname_2_psnames',
        'conv_bbbname_2_psmanames',
        'conv_psname_2_psmaname',
    )

    maname2trims = {
        "SI-Fam:MA-B1B2": None,
        "SI-Fam:MA-QDA": ('SI-01M1:PS-QDA', 'SI-01M2:PS-QDA', 'SI-05M1:PS-QDA',
                          'SI-05M2:PS-QDA', 'SI-09M1:PS-QDA', 'SI-09M2:PS-QDA',
                          'SI-13M1:PS-QDA', 'SI-13M2:PS-QDA', 'SI-17M1:PS-QDA',
                          'SI-17M2:PS-QDA'),
        "SI-14M2:MA-QDB1": None,
        "SI-09M2:MA-SDA0": None,
        "SI-07C4:MA-CH": None,
        "BO-15U:MA-CV": None,
        "TB-03:MA-QF3": None,
        "TS-01:MA-CV-2": None,
        "BO-01D:PM-InjKckr": None,
        "TB-04:PM-InjSept": None,
        "SI-01SA:PM-InjNLKckr": None,
    }

    maname2magfuncs = {
        'SI-Fam:MA-B1B2': {'SI-Fam:PS-B1B2-1': 'dipole',
                           'SI-Fam:PS-B1B2-2': 'dipole'},
        'SI-Fam:MA-QDA': {'SI-Fam:PS-QDA': 'quadrupole'},
        'SI-14M2:MA-QDB1': {'SI-14M2:PS-QDB1': 'quadrupole',
                            'SI-Fam:PS-QDB1': 'quadrupole'},
        'SI-09M2:MA-SDA0': {'SI-09M2:PS-CH': 'corrector-horizontal',
                            'SI-09M2:PS-CV': 'corrector-vertical',
                            'SI-Fam:PS-SDA0': 'sextupole'},
        'SI-07C4:MA-CH': {'SI-07C4:PS-CH': 'corrector-horizontal'},
        'BO-15U:MA-CV': {'BO-15U:PS-CV': 'corrector-vertical'},
        'TB-03:MA-QF3': {'TB-03:PS-QF3': 'quadrupole'},
        'TS-01:MA-CV-2': {'TS-01:PS-CV-2': 'corrector-vertical'},
        'BO-01D:PM-InjKckr': {'BO-01D:PU-InjKckr': 'corrector-horizontal'},
        'TB-04:PM-InjSept': {'TB-04:PU-InjSept': 'corrector-horizontal'},
        'SI-01SA:PM-InjNLKckr':
            {'SI-01SA:PU-InjNLKckr': 'corrector-horizontal'},
    }

    maname2psnames = {
        'SI-Fam:MA-B1B2': ('SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2'),
        'SI-Fam:MA-QDA': ('SI-Fam:PS-QDA',),
        'SI-14M2:MA-QDB1': ('SI-Fam:PS-QDB1', 'SI-14M2:PS-QDB1'),
        'SI-09M2:MA-SDA0': ('SI-Fam:PS-SDA0', 'SI-09M2:PS-CH',
                            'SI-09M2:PS-CV'),
        'SI-07C4:MA-CH': ('SI-07C4:PS-CH',),
        'BO-15U:MA-CV': ('BO-15U:PS-CV',),
        'TB-03:MA-QF3': ('TB-03:PS-QF3',),
        'TS-01:MA-CV-2': ('TS-01:PS-CV-2',),
        'BO-01D:PM-InjKckr': ('BO-01D:PU-InjKckr',),
        'TB-04:PM-InjSept': ('TB-04:PU-InjSept',),
        'SI-01SA:PM-InjNLKckr': ('SI-01SA:PU-InjNLKckr',),
    }

    psname2maname = {
        'SI-Fam:PS-B1B2-1': 'SI-Fam:MA-B1B2',
        'SI-Fam:PS-B1B2-2': 'SI-Fam:MA-B1B2',
        'BO-Fam:PS-B-1': 'BO-Fam:MA-B',
        'BO-Fam:PS-B-2': 'BO-Fam:MA-B',
        'SI-01M1:PS-CH': 'SI-01M1:MA-SDA0',
    }

    psname2psmaname = {
        'SI-Fam:PS-B1B2-1': 'SI-Fam:MA-B1B2',
        'SI-Fam:PS-B1B2-2': 'SI-Fam:MA-B1B2',
        'BO-Fam:PS-B-1': 'BO-Fam:MA-B',
        'BO-Fam:PS-B-1': 'BO-Fam:MA-B',
        'SI-Fam:PS-QDA': 'SI-Fam:MA-QDA',
        'SI-Fam:PS-SDA0': 'SI-Fam:MA-SDA0',
        'SI-01M1:PS-QDA': 'SI-01M1:MA-QDA',
        'SI-01M2:PS-CH': 'SI-01M2:MA-CH',
    }

    def test_public_interface(self):
        """Test class public interface."""
        valid = util.check_public_interface_namespace(
            MASearch, TestMASearch.public_interface)
        self.assertTrue(valid)

    def test_get_manames(self):
        """Test get_manames."""
        manames = MASearch.get_manames()
        self.assertIsInstance(manames, (list, tuple))
        for maname in TestMASearch.maname2trims:
            self.assertIn(maname, manames)
        # check sorted
        # sorted_manames = sorted(manames)
        # self.assertEqual(manames, sorted_manames)
        # with filters
        manames = MASearch.get_manames({'dis': 'PM'})
        self.assertEqual(len(manames), 12)
        for name in manames:
            self.assertIn('PM', name)
        manames = MASearch.get_manames({'sub': '0.M1'})
        self.assertEqual(len(manames), 84)

    def test_conv_maname_2_trims(self):
        """Test conv_maname_2_trims."""
        for ma, trims in TestMASearch.maname2trims.items():
            self.assertEqual(MASearch.conv_maname_2_trims(ma), trims)

    def test_conv_maname_2_magfunc(self):
        """Test conv_maname_2_magfunc."""
        for ma, magfuncs in TestMASearch.maname2magfuncs.items():
            self.assertEqual(MASearch.conv_maname_2_magfunc(ma), magfuncs)

    def test_conv_maname_2_psnames(self):
        """Test conv_maname_2_psnames."""
        for ma, psnames in TestMASearch.maname2psnames.items():
            self.assertEqual(MASearch.conv_maname_2_psnames(ma), psnames)

    # def test_psname_2_maname(self):
    #     """Test psname_2_maname."""
    #     for psname, maname in TestMASearch.psname2maname.items():
    #         self.assertEqual(MASearch.conv_psname_2_maname(psname), maname)

    def test_conv_psname_2_psmaname(self):
        """Test conv_psname_2_psmaname."""
        self.assertIs(MASearch.conv_psname_2_psmaname('INV'), None)
        for psname, maname in TestMASearch.psname2psmaname.items():
            self.assertEqual(MASearch.conv_psname_2_psmaname(psname),
                             maname)

    def test_conv_bbbname_2_psmanames(self):
        """Test conv_bbbname_2_psmanames."""
        self.assertRaises(KeyError, MASearch.conv_bbbname_2_psmanames, '')



"""TB normalized configurations.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return a dict with ramp settings."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict)
    }

    return _dict


_template_dict = {
    'pvs': [
        ['TB-Fam:PS-B:Energy-SP', 0.0, 0.0],  # [GeV]
        ['TB-01:PS-QD1:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-01:PS-QF1:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-02:PS-QD2A:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-02:PS-QF2A:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-02:PS-QD2B:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-02:PS-QF2B:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-03:PS-QD3:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-03:PS-QF3:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-04:PS-QD4:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-04:PS-QF4:KL-SP', 0.0, 0.0],  # [1/m]
        ['TB-01:PS-CH-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-01:PS-CV-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-01:PS-CH-2:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-01:PS-CV-2:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-02:PS-CH-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-02:PS-CV-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-02:PS-CH-2:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-02:PS-CV-2:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-04:PS-CH-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-04:PS-CH-2:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-04:PS-CV-1:Kick-SP', 0.0, 0.0],  # [urad]
        ['TB-04:PS-CV-2:Kick-SP', 0.0, 0.0],  # [urad]
    ]
}
